"""
Results handler classes for Breathe Design simulation outputs.

This module provides convenient interfaces for working with simulation results,
including easy access to KPIs, dynamic data, and integrated plotting capabilities.
"""

import pandas as pd
from typing import List, Dict
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from .plots import plot_dynamic_response, plot_kpis, radar_plot
from .plots import plot_sensitivities as plot_sensitivities_plot
from .utilities import _transform_key, normalize_results
from .dynamic_analysis import extract_dynamic_kpis


class SingleSimulationResults:
    """
    A simple results handler for a single simulation.

    This class provides easy access to KPIs, dynamic data, and plotting capabilities
    for a single simulation result.
    """

    def __init__(self, result: Dict):
        """
        Initialize the SingleSimulationResults handler.

        Args:
            result: Output from run_sim function (single dict)
        """
        self._result = result
        self._input_parameters = result.get("input_parameters", {})

    def _format_kpis(self) -> pd.DataFrame:
        """Format KPIs using the same logic as the API interface."""
        kpis = self._result["KPIs"]
        if "KPInames" in self._result:
            kpi_names = [_transform_key(key) for key in self._result["KPInames"]]
            results_df = pd.DataFrame(kpis)
            results_df["KPI"] = kpi_names
            results_df = results_df[
                ["KPI"] + [col for col in results_df.columns if col != "KPI"]
            ]
            return results_df.set_index("KPI")
        else:
            return pd.DataFrame(kpis)

    @property
    def kpis(self) -> pd.DataFrame:
        """Get formatted KPIs."""
        return self._format_kpis()

    @property
    def dynamic_data(self) -> Dict:
        """Get dynamic data."""
        return self._result.get("dynamicData", {})

    @property
    def design_names(self) -> List[str]:
        """Get list of design names."""
        return list(self.kpis.columns)

    @property
    def baseline(self) -> pd.Series:
        """Get KPIs for the Baseline design if it exists."""
        if "Baseline" in self.kpis.columns:
            return self.kpis["Baseline"]
        elif len(self.kpis.columns) > 0:
            # If no Baseline, return the first design
            return self.kpis.iloc[:, 0]
        else:
            return pd.Series()

    @property
    def capacity(self) -> float:
        """Get the capacity value for the Baseline design (or first design if no Baseline)."""
        baseline_kpis = self.baseline
        if "Capacity [Ah]" in baseline_kpis.index:
            return baseline_kpis["Capacity [Ah]"]
        elif "capacity_Ah" in baseline_kpis.index:
            return baseline_kpis["capacity_Ah"]
        else:
            raise ValueError("Capacity KPI not found in results")

    def get_kpis(self, design_name: str = None) -> pd.DataFrame:
        """Get KPIs for specific design(s)."""
        if design_name is None:
            return self.kpis
        if design_name is not None and design_name not in self.kpis.columns:
            raise ValueError(
                f"Design {design_name} not found. Cannot normalize results."
            )
        return self.kpis[[design_name]]

    def get_normalized_kpis(self, design_name: str = None) -> pd.DataFrame:
        """
        Get normalized KPIs relative to the Baseline design.

        Args:
            design_name (str, optional): Specific design(s) to include. If None, includes all designs.

        Returns:
            pd.DataFrame: Normalized KPIs with Baseline = 1.0 and other designs as ratios

        Raises:
            ValueError: If no Baseline design is found
        """
        kpis = self.get_kpis(design_name)

        # Check if Baseline exists
        if "Baseline" not in kpis.columns:
            raise ValueError("No Baseline design found. Cannot normalize results.")
        if design_name is not None and design_name not in kpis.columns:
            raise ValueError(
                f"Design {design_name} not found. Cannot normalize results."
            )

        return normalize_results(kpis)

    def get_dynamic_data(self, design_name: str = None) -> Dict:
        """Get dynamic data for specific design(s)."""
        if design_name is None:
            return self.dynamic_data
        return (
            {design_name: self.dynamic_data[design_name]}
            if design_name in self.dynamic_data
            else {}
        )

    def compare_designs(
        self, plot_type: str = "relative", designs: List[str] = None
    ) -> go.Figure:
        """Compare designs using KPI plots."""
        kpis = self.kpis
        # Check if we only have Baseline
        if len(kpis.columns) == 1 and "Baseline" in kpis.columns:
            raise ValueError("Only Baseline design found. Cannot compare designs.")
        if designs:
            available_designs = [d for d in designs if d in kpis.columns]
            if available_designs:
                kpis = kpis[available_designs]
        if plot_type == "radar":
            return radar_plot(kpis)
        else:
            return plot_kpis(kpis, plot_type)

    def plot_voltage_response(self, design_name: str = None) -> go.Figure:
        """Plot voltage response for specified design."""
        return self.plot_dynamic_response("voltModel", design_name)

    def plot_dynamic_response(
        self, variable_name: str = "voltModel", design_name: str = None
    ) -> go.Figure:
        """Plot dynamic response for specified variable and design."""
        dynamic_data = self.get_dynamic_data(design_name)
        if not dynamic_data:
            raise ValueError("No dynamic data available for this simulation")
        return plot_dynamic_response(dynamic_data, variable_name)

    def dynamic_kpis(
        self, dynamic_kpi_type: str, design_name: str = None
    ) -> pd.DataFrame:
        """
        Extract dynamic KPIs for specified type and design.

        This method checks if the cycler type matches the requested KPI type
        and extracts the appropriate metrics.

        Args:
            dynamic_kpi_type (str): Type of dynamic KPI to extract ("DCIR" or "RateCap")
            design_name (str, optional): Specific design to analyze. If None, analyzes all designs.

        Returns:
            pd.DataFrame: DataFrame containing the extracted dynamic KPIs

        Raises:
            ValueError: If the cycler type doesn't match the requested KPI type
        """
        # Check dynamic data is not {}
        if not self.dynamic_data:
            raise ValueError("No dynamic data available for this simulation")
        # Check if we have cycler information to validate the request
        if hasattr(self, "_result") and "cycler" in self._result:
            cycler_type = self._result.get("cycler", {}).get("cycle_type", "")
            if dynamic_kpi_type == "DCIR" and cycler_type != "DCIR":
                raise ValueError(
                    f"Requested DCIR KPIs but cycler type is {cycler_type}"
                )
            elif dynamic_kpi_type == "RateCap" and cycler_type not in [
                "CC_CHG",
                "CC_DCH",
            ]:
                raise ValueError(
                    f"Requested RateCap KPIs but cycler type is {cycler_type}"
                )

        dynamic_data = self.get_dynamic_data(design_name)
        return extract_dynamic_kpis(dynamic_kpi_type, dynamic_data)

    def plot_sensitivities(self) -> go.Figure:
        """Plot sensitivity heatmap for the simulation."""
        if "KPInames" not in self._result or "Sensitivities" not in self._result:
            raise ValueError("Sensitivity data not available for this simulation")

        kpi_names = self._result["KPInames"]
        sensitivities = self._result["Sensitivities"]
        return plot_sensitivities_plot(kpi_names, sensitivities)

    def plot_radar(self, designs: List[str] = None) -> go.Figure:
        """
        Create a radar plot comparing designs across KPIs.

        Args:
            designs (List[str], optional): Specific designs to include. If None, includes all designs.

        Returns:
            go.Figure: Radar plot figure
        """
        # Check if we only have Baseline
        if len(self.kpis.columns) == 1 and "Baseline" in self.kpis.columns:
            raise ValueError("Only Baseline design found. Cannot compare designs.")
        kpis = self.get_normalized_kpis(designs)
        return radar_plot(kpis)

    def get_input_parameters(self) -> Dict:
        """Get the original input parameters used for the simulation."""
        return self._input_parameters.copy()

    def get_summary(self) -> Dict:
        """Get a summary of the simulation results."""
        return {
            "is_batch": False,
            "num_simulations": 1,
            "design_names": self.design_names,
            "num_designs": len(self.design_names),
            "input_parameters": self._input_parameters,
        }

    def print_dynamic_variables(self):
        """Print dynamic variables in the dynamic data for the specified design."""
        if not self.dynamic_data:
            raise ValueError("No dynamic data available for this simulation")
        design_name = "Baseline"
        if design_name in self.dynamic_data:
            print("Available variables:")
            for var in self.dynamic_data[design_name].keys():
                print(var)
        else:
            print(f"No dynamic data available for {design_name}")


class BatchSimulationResults:
    """
    A results handler for multiple simulations.

    This class handles a list of simulation results and provides methods
    for comparison and analysis across simulations.
    """

    def __init__(self, results: List[Dict]):
        """
        Initialize the BatchSimulationResults handler.

        Args:
            results: List of outputs from run_sim function
        """
        self._results = results
        self.num_simulations = len(results)

        # Create SingleSimulationResults objects for each result
        self._simulation_results = [
            SingleSimulationResults(result) for result in results
        ]

        # Extract input parameters from each simulation
        self._input_parameters = []
        for sim_result in self._simulation_results:
            self._input_parameters.append(sim_result.get_input_parameters())

    def _calculate_subplot_layout(self) -> tuple[int, int]:
        """
        Calculate optimal rows and columns for subplot layout.

        Returns:
            tuple: (rows, cols) for optimal subplot arrangement
        """
        if self.num_simulations <= 4:
            # For small numbers, prefer horizontal layout
            return 1, self.num_simulations
        elif self.num_simulations <= 6:
            # For medium numbers, try 2 rows
            return 2, (self.num_simulations + 1) // 2
        elif self.num_simulations <= 12:
            # For larger numbers, try 3 rows
            return 3, (self.num_simulations + 2) // 3
        else:
            # For very large numbers, calculate based on square root
            cols = int(self.num_simulations**0.5) + 1
            rows = (self.num_simulations + cols - 1) // cols
            return rows, cols

    def _generate_subplot_titles(self) -> List[str]:
        """
        Generate descriptive subplot titles based on simulation conditions.

        Returns:
            List of titles for each subplot
        """
        titles = []
        for i, sim_result in enumerate(self._simulation_results):
            params = sim_result.get_input_parameters()

            # Create a descriptive title
            title_parts = [f"Sim {i + 1}"]

            # Add key parameters if they exist
            if "initialSoC" in params:
                title_parts.append(f"SOC={params['initialSoC']}")
            if "initialTemperature_degC" in params:
                title_parts.append(f"T={params['initialTemperature_degC']}°C")
            if "ambientTemperature_degC" in params:
                title_parts.append(f"T_amb={params['ambientTemperature_degC']}°C")

            # Join with line breaks for better readability
            titles.append("<br>".join(title_parts))

        return titles

    def __len__(self):
        """Return the number of simulations."""
        return self.num_simulations

    def __getitem__(self, index):
        """Access individual simulation results by index."""
        return self.simulation_results[index]

    def __iter__(self):
        """Iterate over simulation results."""
        for sim_result in self._simulation_results:
            yield sim_result

    @property
    def first(self) -> "SingleSimulationResults":
        """Get the first simulation result."""
        return self._simulation_results[0]

    @property
    def baseline(self) -> pd.Series:
        """Get KPIs for the Baseline design from the first simulation."""
        return self.first.baseline

    @property
    def capacity(self) -> float:
        """Get the capacity value for the Baseline design from the first simulation."""
        return self.first.capacity

    @property
    def design_names(self) -> List[str]:
        """Get the design names for the first simulation."""
        return self.first.design_names

    @property
    def dynamic_data(self) -> List[Dict]:
        """Get dynamic data for all simulations."""
        return [sim.dynamic_data for sim in self._simulation_results]

    def get_kpis(self, designs: List[str] = None) -> List[pd.DataFrame]:
        """Get KPIs for all the first simulation."""
        return self.first.get_kpis(designs)

    def get_normalized_kpis(self, designs: List[str] = None) -> List[pd.DataFrame]:
        """
        Get normalized KPIs for all simulations relative to their Baseline designs.

        Args:
            designs (List[str], optional): Specific design(s) to include. If None, includes all designs.

        Returns:
            pd.DataFrame: List of normalized KPIs for each simulation

        Raises:
            ValueError: If no Baseline design is found in any simulation
        """
        return self.first.get_normalized_kpis(designs)

    def get_dynamic_data(self, design_name: str = None) -> List[Dict]:
        """Get dynamic data for all simulations."""
        return [sim.get_dynamic_data(design_name) for sim in self._simulation_results]

    def compare_designs(
        self, plot_type: str = "relative", designs: List[str] = None
    ) -> go.Figure:
        """Compare designs across all simulations using subplots."""
        if plot_type == "radar":
            raise ValueError("Radar plots not supported for batch simulations")

        # Calculate optimal rows and columns for better layout
        rows, cols = self._calculate_subplot_layout()

        # Get KPIs for all simulations
        kpis_list = self.get_kpis(designs)
        # Check if we only have Baseline
        if len(kpis_list[0].columns) == 1 and "Baseline" in kpis_list[0].columns:
            raise ValueError("Only Baseline design found. Cannot compare designs.")
        # Create subplots with calculated layout
        subplot_titles = self._generate_subplot_titles()
        fig = make_subplots(rows=rows, cols=cols, subplot_titles=subplot_titles)

        for i, kpis in enumerate(kpis_list):
            if not kpis.empty:
                temp_fig = plot_kpis(kpis, plot_type)
                # Calculate row and column position
                row = (i // cols) + 1
                col = (i % cols) + 1
                for trace in temp_fig.data:
                    fig.add_trace(trace, row=row, col=col)

        fig.update_layout(
            title=f"Design Comparison - {plot_type.title()}",
            height=300 * rows,
            # Add margin to prevent title overlap
            margin=dict(t=100),
            # Reduce subtitle text size
            font=dict(size=10),
        )
        return fig

    def plot_dynamic_response(
        self, variable_name: str = "voltModel", design_name: str = None
    ) -> go.Figure:
        """Plot dynamic response across all simulations using subplots."""
        # Check if dynamic data is available for any simulation
        if not self._simulation_results[0].dynamic_data:
            raise ValueError("No dynamic data available.")

        # Calculate optimal rows and columns for better layout
        rows, cols = self._calculate_subplot_layout()
        subplot_titles = self._generate_subplot_titles()
        fig = make_subplots(rows=rows, cols=cols, subplot_titles=subplot_titles)

        for i, sim_result in enumerate(self._simulation_results):
            temp_fig = sim_result.plot_dynamic_response(variable_name, design_name)
            # Calculate row and column position
            row = (i // cols) + 1
            col = (i % cols) + 1
            for trace in temp_fig.data:
                fig.add_trace(trace, row=row, col=col)

        fig.update_layout(height=300 * rows)
        return fig

    def plot_voltage_response(self, design_name: str = None) -> go.Figure:
        """Plot voltage response across all simulations using subplots."""
        return self.plot_dynamic_response("voltModel", design_name)

    def dynamic_kpis(
        self, dynamic_kpi_type: str, design_name: str = None
    ) -> List[pd.DataFrame]:
        """
        Extract dynamic KPIs for all simulations.

        This method checks if the cycler type matches the requested KPI type
        and extracts the appropriate metrics for each simulation.

        Args:
            dynamic_kpi_type (str): Type of dynamic KPI to extract ("DCIR" or "RateCap")
            design_name (str, optional): Specific design to analyze. If None, analyzes all designs.

        Returns:
            List[pd.DataFrame]: List of DataFrames containing the extracted dynamic KPIs for each simulation

        Raises:
            ValueError: If the cycler type doesn't match the requested KPI type
        """
        return [
            sim.dynamic_kpis(dynamic_kpi_type, design_name)
            for sim in self._simulation_results
        ]

    def plot_sensitivities(self) -> go.Figure:
        """Plot sensitivity heatmap for the first simulation in the batch."""
        return self.first.plot_sensitivities()

    def plot_radar(self, designs: List[str] = None) -> go.Figure:
        """
        Create a radar plot comparing designs across KPIs for the first simulation.

        Args:
            designs (List[str], optional): Specific designs to include. If None, includes all designs.

        Returns:
            go.Figure: Radar plot figure
        """
        return self.first.plot_radar(designs)

    def get_input_parameters(self) -> List[Dict]:
        """Get the input parameters for all simulations."""
        return self._input_parameters.copy()

    def get_summary(self) -> Dict:
        """Get a summary of all simulation results."""
        return {
            "is_batch": True,
            "num_simulations": self.num_simulations,
            "design_names": self.first.design_names if self._simulation_results else [],
            "num_designs": len(self.first.design_names)
            if self._simulation_results
            else 0,
            "input_parameters": self._input_parameters,
        }

    def print_dynamic_variables(self):
        """Print dynamic variables in the dynamic data for the first simulation."""
        self.first.print_dynamic_variables()



"""
This script is run by mkdocs to insert variables into the mkdocs environment
that can then be accessed inside the Markdown documents
e.g.
{{ __version__ }}
"""

import os

_this_file_dir = os.path.dirname(os.path.abspath(__file__))


def define_env(env):
    # extract the __version__ number of the module
    # don't try and import it because we can't be sure what context this will be run in
    breathe_design_init_file_path = os.path.join(
        _this_file_dir, "breathe_design", "__init__.py"
    )
    with open(breathe_design_init_file_path, "r") as f:
        lines = f.readlines()
    version = None
    for line in lines:
        if "__version__" not in line:
            continue
        parts = line.split(" = ")
        version = parts[1].strip().strip('"')
    print(version)
    env.variables["breathe_design_version"] = version



import copy
import json
import os
from typing import Dict, List, Any, Optional


class ParameterValidator:
    """
    Unified validator for design and format parameters used in the breathe design process.
    Ensures parameters are within acceptable bounds before API calls to prevent errors.
    """

    def __init__(self):
        """
        Initialize the ParameterValidator with parameter bounds from JSON file.
        """
        self._parameter_bounds = self._load_parameter_bounds()

    def _load_parameter_bounds(self) -> dict:
        """
        Load parameter bounds from the unified JSON configuration file.

        Returns:
            dict: Dictionary containing parameter bounds with [min, max] values.
        """
        current_dir = os.path.dirname(os.path.abspath(__file__))
        json_path = os.path.join(current_dir, "parameter_bounds.json")

        try:
            with open(json_path, "r") as f:
                bounds = json.load(f)
            return bounds
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Could not find parameter bounds file at {json_path}. "
                "Please ensure parameter_bounds.json exists in the breathe_design package directory."
            )
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in parameter bounds file: {e}")

    def validate_design(self, design: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate a single design dictionary against parameter bounds.

        Args:
            design (Dict[str, Any]): Dictionary containing design parameters.

        Returns:
            Dict[str, Any]: The validated design dictionary.

        Raises:
            ValueError: If any parameter is out of bounds or invalid.
        """
        return self._validate_parameters(
            design, self._parameter_bounds, "design", ["designName"]
        )

    def validate_designs(self, designs: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Validate a list of design dictionaries.

        Args:
            designs (List[Dict[str, Any]]): List of design dictionaries.

        Returns:
            List[Dict[str, Any]]: List of validated design dictionaries.

        Raises:
            ValueError: If any design contains invalid parameters.
        """
        return self._validate_parameters_list(designs, self.validate_design, "Design")

    def validate_format(self, format_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate a single format dictionary against parameter bounds and constraints.

        Args:
            format_dict (Dict[str, Any]): Dictionary containing format parameters.

        Returns:
            Dict[str, Any]: The validated format dictionary.

        Raises:
            ValueError: If any parameter is out of bounds or invalid.
        """
        # Import here to avoid circular imports
        from .format import validate_material, validate_assembly_type

        if not isinstance(format_dict, dict):
            raise ValueError("Format must be a dictionary")

        # Make a copy to avoid modifying the original
        validated_format = copy.deepcopy(format_dict)

        # Validate required fields
        if "shape" not in validated_format:
            raise ValueError("Format must specify a 'shape' field")

        if "material" not in validated_format:
            raise ValueError("Format must specify a 'material' field")

        # Validate material and assembly type
        validate_material(validated_format["material"])
        if "assemblyType" in validated_format:
            validate_assembly_type(validated_format["assemblyType"])

        # Skip format-specific fields during parameter validation
        skip_fields = ["shape", "name", "material", "assemblyType"]

        # Validate numeric parameters
        validated_format = self._validate_parameters(
            validated_format, self._parameter_bounds, "format", skip_fields
        )

        # Shape-specific validations
        if validated_format["shape"] == "cylinder":
            self._validate_cylinder_constraints(validated_format)
        elif validated_format["shape"] == "cuboid":
            self._validate_cuboid_constraints(validated_format)

        return validated_format

    def validate_formats(self, formats: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Validate a list of format dictionaries.

        Args:
            formats (List[Dict[str, Any]]): List of format dictionaries.

        Returns:
            List[Dict[str, Any]]: List of validated format dictionaries.

        Raises:
            ValueError: If any format contains invalid parameters.
        """
        return self._validate_parameters_list(formats, self.validate_format, "Format")

    def _validate_parameters(
        self,
        param_dict: Dict[str, Any],
        bounds: Dict[str, List],
        param_type: str,
        skip_fields: List[str],
    ) -> Dict[str, Any]:
        """
        Core parameter validation logic.

        Args:
            param_dict (Dict[str, Any]): Dictionary containing parameters to validate.
            bounds (Dict[str, List]): Parameter bounds to validate against.
            param_type (str): Type of parameters for error messages ("design" or "format").
            skip_fields (List[str]): Field names to skip during validation.

        Returns:
            Dict[str, Any]: The validated parameter dictionary.
        """
        if not isinstance(param_dict, dict):
            raise ValueError(f"{param_type.capitalize()} must be a dictionary")

        # Make a copy to avoid modifying the original
        validated_params = copy.deepcopy(param_dict)

        # Check each parameter
        for param_name, param_value in validated_params.items():
            # Skip non-parameter fields
            if param_name in skip_fields:
                continue

            if param_name in bounds:
                self._validate_single_parameter(
                    param_name, param_value, bounds[param_name], param_type
                )
            # Allow unknown parameters for future extensibility

        return validated_params

    def _validate_parameters_list(
        self, param_list: List[Dict[str, Any]], validator_func, param_type_name: str
    ) -> List[Dict[str, Any]]:
        """
        Validate a list of parameter dictionaries.

        Args:
            param_list (List[Dict[str, Any]]): List of parameter dictionaries.
            validator_func: Function to validate individual dictionaries.
            param_type_name (str): Name for error messages.

        Returns:
            List[Dict[str, Any]]: List of validated parameter dictionaries.
        """
        if not isinstance(param_list, list):
            raise ValueError(f"{param_type_name}s must be a list")

        validated_list = []
        for i, param_dict in enumerate(param_list):
            try:
                validated_param = validator_func(param_dict)
                validated_list.append(validated_param)
            except ValueError as e:
                raise ValueError(f"{param_type_name} {i} validation failed: {e}")

        return validated_list

    def _validate_single_parameter(
        self, param_name: str, param_value: Any, bounds: List, param_type: str
    ) -> None:
        """
        Validate a single parameter against its bounds.

        Args:
            param_name (str): Name of the parameter.
            param_value (Any): Value of the parameter.
            bounds (List): [min, max] bounds for the parameter.
            param_type (str): Type of parameter for error messages.

        Raises:
            ValueError: If the parameter value is invalid or out of bounds.
        """
        if param_value is None:
            raise ValueError(
                f"{param_type.capitalize()} parameter '{param_name}' cannot be None"
            )

        # Check if parameter is numeric
        if not isinstance(param_value, (int, float)):
            raise ValueError(
                f"{param_type.capitalize()} parameter '{param_name}' must be numeric, got {type(param_value)}"
            )

        min_val, max_val = bounds

        # Check minimum bound (if not None)
        if min_val is not None and param_value < min_val:
            raise ValueError(
                f"{param_type.capitalize()} parameter '{param_name}' value {param_value} is below minimum allowed value {min_val}"
            )

        # Check maximum bound (if not None)
        if max_val is not None and param_value > max_val:
            raise ValueError(
                f"{param_type.capitalize()} parameter '{param_name}' value {param_value} is above maximum allowed value {max_val}"
            )

    def _validate_cylinder_constraints(self, format_dict: Dict[str, Any]) -> None:
        """
        Validate cylinder-specific constraints.

        Args:
            format_dict (Dict[str, Any]): Cylinder format dictionary.
        """
        if "diameter_mm" in format_dict and "innerDiameter_mm" in format_dict:
            diameter = format_dict["diameter_mm"]
            inner_diameter = format_dict["innerDiameter_mm"]

            if inner_diameter >= diameter:
                raise ValueError(
                    f"Inner diameter ({inner_diameter}) must be less than outer diameter ({diameter})"
                )

    def _validate_cuboid_constraints(self, format_dict: Dict[str, Any]) -> None:
        """
        Validate cuboid-specific constraints.

        Args:
            format_dict (Dict[str, Any]): Cuboid format dictionary.
        """
        # Add any cuboid-specific validation logic here if needed
        pass

    def format_parameters(self, param_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Format parameters for API consistency (e.g., ensuring proper types).

        Args:
            param_dict (Dict[str, Any]): Dictionary containing parameters.

        Returns:
            Dict[str, Any]: Formatted parameter dictionary.
        """
        formatted_params = copy.deepcopy(param_dict)

        # Use the unified bounds

        # Ensure numeric parameters are float type for API consistency
        for param_name, param_value in formatted_params.items():
            if param_name in [
                "designName",
                "shape",
                "name",
                "material",
                "assemblyType",
            ]:
                continue

            if param_name in self._parameter_bounds and param_value is not None:
                if isinstance(param_value, (int, float)):
                    formatted_params[param_name] = float(param_value)

        return formatted_params

    def validate_and_format_designs(
        self, designs: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """
        Validate and format a list of designs in one step.

        Args:
            designs (List[Dict[str, Any]]): List of design dictionaries.

        Returns:
            List[Dict[str, Any]]: List of validated and formatted design dictionaries.
        """
        validated_designs = self.validate_designs(designs)
        formatted_designs = [
            self.format_parameters(design) for design in validated_designs
        ]
        return formatted_designs

    def validate_and_format_formats(
        self, formats: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """
        Validate and format a list of formats in one step.

        Args:
            formats (List[Dict[str, Any]]): List of format dictionaries.

        Returns:
            List[Dict[str, Any]]: List of validated and formatted format dictionaries.
        """
        validated_formats = self.validate_formats(formats)
        formatted_formats = [
            self.format_parameters(format_dict) for format_dict in validated_formats
        ]
        return formatted_formats

    def get_parameter_bounds(self, param_name: Optional[str] = None) -> Dict[str, List]:
        """
        Get parameter bounds for specific parameters or all parameters.

        Args:
            param_name (Optional[str]): Name of specific parameter, or None for all parameters.

        Returns:
            Dict[str, List]: Parameter bounds as {parameter_name: [min, max]}.
        """
        if param_name is None:
            return copy.deepcopy(self._parameter_bounds)

        if param_name not in self._parameter_bounds:
            raise ValueError(f"Unknown parameter: {param_name}")

        return {param_name: copy.deepcopy(self._parameter_bounds[param_name])}


# Create singleton instance for convenience
parameter_validator = ParameterValidator()



import copy
import json
import os


class Cycler:
    """
    Represents a battery cycler protocol used in the breathe design process for dynamic analysis. Has methods to generate different battery cycler protocols in the correct format.
    """

    def __init__(self, selected_unit: str, cell_capacity: float):
        """
        Initialize a Cycler object with the selected unit and cell capacity. The default protocol is set to CC_CHG.

        Parameters:
            selected_unit (str): The unit of current (A or C) used in the cycler.
            cell_capacity (float): The capacity of the battery cell in Ah.
        """
        self._selected_unit = selected_unit
        self._cell_capacity = cell_capacity
        self._default_cycler = {
            "cycle_type": None,
            "selected_unit": selected_unit,
            "control_parameters": {},
        }
        self._protocols = self._load_protocols()

    def _load_protocols(self) -> dict:
        """
        Load cycling protocols from the JSON configuration file.

        Returns:
            dict: Dictionary containing protocol definitions with parameter limits.
        """
        # Get the directory where this module is located
        current_dir = os.path.dirname(os.path.abspath(__file__))
        json_path = os.path.join(current_dir, "cycling_protocols.json")

        try:
            with open(json_path, "r") as f:
                protocols = json.load(f)
            return protocols
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Could not find cycling protocols file at {json_path}. "
                "Please ensure cycling_protocols.json exists in the breathe_design package directory."
            )
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in cycling protocols file: {e}")

    def cc_chg(self, I_chg: float, V_max: float) -> dict:
        """
        Create a Constant Current Charge cycler protocol.

        Parameters:
            I_chg (float): The charging current in [A || C]. Required.
            V_max (float): The maximum charging voltage in V. Required.

        Returns:
            (dict): representing the CC-CHG cycling protocol.

        """
        cycler = copy.deepcopy(self._default_cycler)
        cycler["cycle_type"] = "CC_CHG"
        cycler["control_parameters"]["I_chg"] = I_chg
        cycler["control_parameters"]["V_max"] = V_max
        return self._build_cycler_dict(cycler)

    def cc_dch(self, I_dch: float, V_min: float) -> dict:
        """
        Create a Constant Current Discharge cycler protocol.

        Args:
            I_dch (float): The discharging current in [A || C]. Required.
            V_min (float): The minimum discharging voltage in V. Required.

        Returns:
            (dict): representing the CC-DCH cycling protocol.
        """
        cycler = copy.deepcopy(self._default_cycler)
        cycler["cycle_type"] = "CC_DCH"
        cycler["control_parameters"]["I_dch"] = I_dch
        cycler["control_parameters"]["V_min"] = V_min
        return self._build_cycler_dict(cycler)

    def dcir(
        self,
        I_dch: float,
        t_dur: float,
        t_rest_before: float,
        t_rest_after: float,
        v_min: float,
        v_max: float,
    ) -> dict:
        """
        Create a Direct Current Internal Resistance (DCIR) cycler protocol.

        Parameters:
            I_dch (float): The discharging current in [A || C]. Required.
            t_dur (float): The duration of the discharge in seconds. Required.
            t_rest_before (float): The rest time before the discharge in seconds. Required.
            t_rest_after (float): The rest time after the discharge in seconds. Required.
            v_min (float): The minimum voltage in V. Required.
            v_max (float): The maximum voltage in V. Required.

        Returns:
            (dict): representing the DCIR cycling protocol.
        """
        cycler = copy.deepcopy(self._default_cycler)
        cycler["cycle_type"] = "DCIR"
        cycler["control_parameters"]["I_dch"] = I_dch
        cycler["control_parameters"]["t_dur"] = t_dur
        cycler["control_parameters"]["t_rest_before"] = t_rest_before
        cycler["control_parameters"]["t_rest_after"] = t_rest_after
        cycler["control_parameters"]["v_min"] = v_min
        cycler["control_parameters"]["v_max"] = v_max
        return self._build_cycler_dict(cycler)

    def cccv(
        self,
        I_chg: float,
        I_dch: float,
        I_cut: float,
        V_max: float,
        V_min: float,
        t_restC: float,
        t_restD: float,
    ) -> dict:
        """
        Create a Constant Current Constant Voltage (CCCV) cycler protocol.

        Parameters:
            I_chg (float): The charging current in [A || C]. Required.
            I_dch (float): The discharging current in [A || C]. Required.
            I_cut (float): The cut-off current in [A || C]. Required.
            V_max (float): The maximum charging voltage in V. Required.
            V_min (float): The minimum discharging voltage in V. Required.
            t_restC (float): The rest time (in seconds) after charging. Required.
            t_restD (float): The rest time (in seconds) after discharging. Required.

        Returns:
            (dict): representing the CCCV cycling protocol.
        """
        cycler = copy.deepcopy(self._default_cycler)
        cycler["cycle_type"] = "CCCV"
        cycler["control_parameters"]["I_chg"] = I_chg
        cycler["control_parameters"]["I_dch"] = I_dch
        cycler["control_parameters"]["I_cut"] = I_cut
        cycler["control_parameters"]["V_max"] = V_max
        cycler["control_parameters"]["V_min"] = V_min
        cycler["control_parameters"]["t_restC"] = t_restC
        cycler["control_parameters"]["t_restD"] = t_restD
        return self._build_cycler_dict(cycler)

    def rate_cap(
        self,
        I_chg: float,
        I_dch: float,
        I_cut: float,
        V_max: float,
        V_min: float,
        t_restC: float,
    ) -> dict:
        """
        Create a Rate Capacity cycler protocol.

        Parameters:
            I_chg (float): The charging current in [A || C]. Required.
            I_dch (float): The discharging current in [A || C]. Required.
            I_cut (float): The cut-off current in [A || C]. Required.
            V_max (float): The maximum charging voltage in V. Required.
            V_min (float): The minimum discharging voltage in V. Required.
            t_restC (float): The rest time (in seconds) after charging. Required.

        Returns:
            (dict): representing the RateCap cycling protocol.
        """
        cycler = copy.deepcopy(self._default_cycler)
        cycler["cycle_type"] = "RateCap"
        cycler["control_parameters"]["I_chg"] = I_chg
        cycler["control_parameters"]["I_dch"] = I_dch
        cycler["control_parameters"]["I_cut"] = I_cut
        cycler["control_parameters"]["V_max"] = V_max
        cycler["control_parameters"]["V_min"] = V_min
        cycler["control_parameters"]["t_restC"] = t_restC
        return self._build_cycler_dict(cycler)

    def _build_cycler_dict(self, cycler: dict) -> dict:
        if cycler["cycle_type"] not in self._protocols:
            raise ValueError(f"Invalid cycle type: {cycler['cycle_type']}")

        protocol_params = self._protocols[cycler["cycle_type"]]

        for param in protocol_params:
            if param not in cycler["control_parameters"]:
                raise ValueError(f"Missing required parameter: {param}")
            # copy value for checking
            value = copy.deepcopy(cycler["control_parameters"][param])
            if value is None:
                raise ValueError(f"Missing required parameter: {param}")
            if self._selected_unit == "C":
                if param in ["I_chg", "I_dch", "I_cut"]:
                    value *= (
                        self._cell_capacity
                    )  # Convert from multiples of the cell's capacity (C) to Amps

            min_val, max_val = protocol_params[param]
            if value < min_val or value > max_val:
                raise ValueError(
                    f"Invalid value for {param}. Must be between {min_val} and {max_val}."
                )

        return cycler



#!/usr/bin/env python3
"""
Script to copy files and folders from this repo to a remote repo and create requirements.txt
"""

import sys
import shutil
import argparse
from pathlib import Path
import re


def get_version_from_init():
    """Extract version from breathe_design/__init__.py"""
    init_file = Path(__file__).parent.parent / "breathe_design" / "__init__.py"

    if not init_file.exists():
        raise FileNotFoundError(f"Could not find __init__.py at {init_file}")

    with open(init_file, "r") as f:
        content = f.read()

    # Look for __version__ = "x.y.z" pattern
    version_match = re.search(r'__version__\s*=\s*["\']([^"\']+)["\']', content)
    if not version_match:
        raise ValueError("Could not find __version__ in __init__.py")

    return version_match.group(1)


def get_files_to_copy():
    """
    Define the list of files/folders to copy and their destination paths.
    Returns a list of tuples: (source_path, destination_path)
    """
    return [
        # Documentation and examples
        ("docs/examples", "examples"),
        ("docs/installation.md", "docs/installation.md"),
        ("docs/create_a_sim.md", "docs/create_a_sim.md"),
        ("README.md", "README.md"),
        ("docs/PUBLIC_CHANGELOG.md", "CHANGELOG.md"),
        ("TROUBLESHOOTING.md", "TROUBLESHOOTING.md"),
        # Setup script
        ("public_release_scripts/setupEnvironment.ps1", "setupEnvironment.ps1"),
    ]


def clean_destination_repo(dest_repo):
    """
    Delete all contents of the destination repository except git-related files.
    Preserves: .git, .gitignore, .gitmodules, .gitattributes
    """
    dest_path = Path(dest_repo)

    if not dest_path.exists():
        print(f"Destination repo doesn't exist yet: {dest_path}")
        return [], []

    deleted_items = []
    failed_items = []

    try:
        # Get all items in the destination directory
        items = list(dest_path.iterdir())

        if not items:
            print("Destination repo is already empty")
            return [], []

        print(f"Cleaning destination repo: {dest_path}")

        for item in items:
            try:
                # Skip git-related files and folders to preserve repository history
                if item.name in [".git", ".gitignore", ".gitmodules", ".gitattributes"]:
                    print(f"⚠ Skipping {item.name} (preserving git configuration)")
                    continue

                if item.is_file():
                    item.unlink()
                    print(f"✓ Deleted file: {item.name}")
                    deleted_items.append(item.name)
                elif item.is_dir():
                    shutil.rmtree(item)
                    print(f"✓ Deleted folder: {item.name}")
                    deleted_items.append(item.name)
            except Exception as e:
                print(f"✗ Error deleting {item.name}: {e}")
                failed_items.append(item.name)

        if failed_items:
            print(f"⚠ Failed to delete {len(failed_items)} items: {failed_items}")

    except Exception as e:
        print(f"✗ Error accessing destination repo: {e}")
        failed_items.append(str(dest_path))

    return deleted_items, failed_items


def copy_files_and_folders(source_repo, dest_repo, files_to_copy):
    """Copy files and folders from source to destination repo"""
    source_path = Path(source_repo)
    dest_path = Path(dest_repo)

    # Create destination directory if it doesn't exist
    dest_path.mkdir(parents=True, exist_ok=True)

    copied_files = []
    failed_files = []

    for source_rel, dest_rel in files_to_copy:
        source_full = source_path / source_rel
        dest_full = dest_path / dest_rel

        try:
            if source_full.exists():
                # Create parent directories if they don't exist
                dest_full.parent.mkdir(parents=True, exist_ok=True)

                if source_full.is_file():
                    shutil.copy2(source_full, dest_full)
                    print(f"✓ Copied file: {source_rel} -> {dest_rel}")
                    copied_files.append(dest_rel)
                elif source_full.is_dir():
                    # Remove destination if it exists and copy the directory
                    if dest_full.exists():
                        shutil.rmtree(dest_full)
                    shutil.copytree(source_full, dest_full)
                    print(f"✓ Copied folder: {source_rel} -> {dest_rel}")
                    copied_files.append(dest_rel)
            else:
                print(f"⚠ Warning: Source path does not exist: {source_rel}")
                failed_files.append(source_rel)

        except Exception as e:
            print(f"✗ Error copying {source_rel}: {e}")
            failed_files.append(source_rel)

    return copied_files, failed_files


def create_requirements_txt(dest_repo, version):
    """Create requirements.txt file in the destination repo"""
    dest_path = Path(dest_repo)
    requirements_file = dest_path / "requirements.txt"

    try:
        with open(requirements_file, "w") as f:
            f.write(f"breathe_design >= {version}\n")

        print(f"✓ Created requirements.txt with breathe_design >= {version}")
        return True

    except Exception as e:
        print(f"✗ Error creating requirements.txt: {e}")
        return False


def main():
    parser = argparse.ArgumentParser(
        description="Copy files from this repo to a remote repo and create requirements.txt"
    )
    parser.add_argument(
        "repo_path",
        nargs="?",
        default="D:/Git/breathe_design_examples",
        help="Path to the destination repository (default: D:/Git/breathe_design_examples)",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Show what would be copied without actually copying",
    )

    args = parser.parse_args()

    # Get current repo path (go up one level from public_release_scripts)
    current_repo = Path(__file__).parent.parent
    dest_repo = Path(args.repo_path)

    print(f"Source repo: {current_repo}")
    print(f"Destination repo: {dest_repo}")
    print()

    try:
        # Get version from __init__.py
        version = get_version_from_init()
        print(f"Found breathe_design version: {version}")
        print()

        # Get list of files to copy
        files_to_copy = get_files_to_copy()

        if args.dry_run:
            print("DRY RUN - What would happen:")
            print()

            # Show what would be deleted
            if dest_repo.exists():
                items = list(dest_repo.iterdir())
                if items:
                    print("Would DELETE all existing files/folders:")
                    git_files_found = []
                    git_protected_files = [
                        ".git",
                        ".gitignore",
                        ".gitmodules",
                        ".gitattributes",
                    ]

                    for item in items:
                        if item.name in git_protected_files:
                            git_files_found.append(item.name)
                            continue  # Skip git files in display
                        file_type = "folder" if item.is_dir() else "file"
                        print(f"  ✗ {item.name} ({file_type})")

                    if git_files_found:
                        print(
                            f"  ⚠ Git files would be PRESERVED: {', '.join(git_files_found)}"
                        )

                    if not any(item.name not in git_protected_files for item in items):
                        print("  (Only git-related files present - would be preserved)")
                else:
                    print("Destination repo is empty - nothing to delete")
            else:
                print("Destination repo doesn't exist yet - would create it")
            print()

            # Show what would be copied
            print("Would COPY these files:")
            for source_rel, dest_rel in files_to_copy:
                source_full = current_repo / source_rel
                if source_full.exists():
                    file_type = "folder" if source_full.is_dir() else "file"
                    print(f"  ✓ {source_rel} -> {dest_rel} ({file_type})")
                else:
                    print(f"  ⚠ {source_rel} -> {dest_rel} (MISSING)")
            print()
            print(f"Would CREATE requirements.txt with: breathe_design >= {version}")
            return

        # Clean destination repo first
        print("Cleaning destination repository...")
        deleted_items, delete_failed = clean_destination_repo(dest_repo)
        print()

        # Copy files and folders
        print("Copying files and folders...")
        copied_files, failed_files = copy_files_and_folders(
            current_repo, dest_repo, files_to_copy
        )
        print()

        # Create requirements.txt
        print("Creating requirements.txt...")
        requirements_success = create_requirements_txt(dest_repo, version)
        print()

        # Summary
        print("=" * 50)
        print("SUMMARY")
        print("=" * 50)
        print(f"Deleted {len(deleted_items)} existing items")
        if delete_failed:
            print(f"Failed to delete {len(delete_failed)} items: {delete_failed}")
        print(f"Successfully copied {len(copied_files)} items")
        if failed_files:
            print(f"Failed to copy {len(failed_files)} items: {failed_files}")
        print(f"Requirements.txt created: {'✓' if requirements_success else '✗'}")
        print(f"Destination: {dest_repo}")

    except Exception as e:
        print(f"✗ Error: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



import json
from pathlib import Path
from breathe_design import api_interface as api

import nbformat
import pytest
from nbclient import NotebookClient

# ---- Collect all notebooks under docs/examples ----
NOTEBOOK_DIR = Path("docs") / "examples"
ALL_NOTEBOOKS = sorted(p for p in NOTEBOOK_DIR.rglob("*.ipynb"))


def test__show_service_version(use_m2m_auth):
    version = api.get_service_version()
    print(f"Service version is {version}")


@pytest.mark.parametrize(
    "notebook_path", ALL_NOTEBOOKS, ids=lambda p: str(p.relative_to(NOTEBOOK_DIR))
)
def test_run_notebook_without_errors(notebook_path, use_m2m_auth):
    """
    Execute the notebook in a clean Jupyter kernel and fail if any cell errors.
    We also inject a setup cell that reproduces the same patching as the fixture,
    but *inside the kernel* the notebook runs in.
    """
    # Grab the token value from your fixture so we can mirror the patch in the kernel.
    # Your fixture yields the mock, whose return_value is the token string.
    token_value = use_m2m_auth.return_value

    # Load the notebook
    nb = nbformat.read(notebook_path, as_version=4)

    # Inject a setup cell at the very top that applies the same patch logic
    # NOTE: the patch needs to happen in the *kernel process* executing the notebook.
    setup_code = f"""
from unittest.mock import patch
from breathe_design.device_flow_auth import device_auth
# Start a patcher that lasts for the entire notebook run.
__pytest_nb_patcher = patch.object(device_auth, "get_token", return_value={json.dumps(token_value)})
__pytest_nb_patcher.start()
"""
    nb.cells.insert(0, nbformat.v4.new_code_cell(setup_code))

    # Execute with nbclient; set working dir to the notebook's folder so relative paths work
    client = NotebookClient(
        nb,
        kernel_name="python3",
        timeout=600,
        resources={"metadata": {"path": str(notebook_path.parent)}},
        allow_errors=False,  # raise on first error
    )

    # Run! If any cell errors, nbclient will raise and pytest will fail.
    client.execute()



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
import httpx
from functools import lru_cache
import pytest
from unittest.mock import patch
import breathe_design.device_flow_auth
import dotenv

dotenv.load_dotenv()


class Auth0Config(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", extra="ignore", frozen=True)

    service_iss: str = Field(validation_alias="AUTH0_SERVICE_ISS")
    audience: str = Field(validation_alias="AUTH0_AUDIENCE")
    client_id: str = Field(validation_alias="AUTH0_CLIENT_ID")
    client_secret: str = Field(validation_alias="AUTH0_CLIENT_SECRET")


@lru_cache
def _get_access_token() -> str:
    """Request an access token from Auth0 using the Client Secret flow

    Raises:
        RuntimeError: Raised if endpoint returns non-200

    Returns:
        str: The access token string
    """
    auth0_config = Auth0Config()
    payload = {
        "client_id": auth0_config.client_id,
        "client_secret": auth0_config.client_secret,
        "audience": auth0_config.audience,
        "grant_type": "client_credentials",
    }
    headers = {"content-type": "application/x-www-form-urlencoded"}

    resp = httpx.post(
        f"{auth0_config.service_iss}/oauth/token", data=payload, headers=headers
    )

    if resp.status_code != 200:
        raise RuntimeError(
            f"Token request from Auth0 failed with code {resp.status_code}: {resp.reason_phrase}"
        )

    return resp.json()["access_token"]


@pytest.fixture
def use_m2m_auth():
    token = _get_access_token()
    with patch.object(
        breathe_design.device_flow_auth.device_auth, "get_token", return_value=token
    ) as mock_method:
        yield mock_method



import logging
from typing import Literal, Union, List
import requests
import tempfile
import pandas as pd
from .utilities import _transform_key
from .results import SingleSimulationResults, BatchSimulationResults
from concurrent.futures import ThreadPoolExecutor
from .device_flow_auth import device_auth
from packaging.version import Version
import json
from .api_utils import (
    BreatheException,
    make_design_names_map,
    map_design_names__human_to_machine,
    map_fields_in_place,
    convert_none_strings_to_none,
)
from .parameter_validator import parameter_validator
import os
import dotenv
from .__init__ import __version__
import re

dotenv.load_dotenv()

_logger = logging.getLogger(__name__)

temp_dir = tempfile.mkdtemp(prefix="myapp_")


class BreatheDesignModel:
    """
    A class for interacting with the Breathe Design API.
    """

    def __init__(self):
        self._auth = device_auth
        if "API_URL" in os.environ:
            self._api_url = os.environ["API_URL"]
            _logger.info(f"API_URL set from environment to {self._api_url}")
        else:
            self._api_url = (
                "https://breathe-platform-backend-service-prod.azurewebsites.net/api/v1"
            )

    def ensure_logged_in(self):
        """
        Checks if a valid token is available and refreshes it if necessary.
        """
        device_auth.get_token()

    def _make_api_call(
        self, verb: Literal["POST", "GET"], endpoint: str, payload: dict | None
    ) -> dict:
        token = device_auth.get_token()
        headers = {
            "accept": "application/json",
            "Authorization": f"Bearer {token}",
        }
        if payload is not None:
            headers["Content-Type"] = "application/json"

        try:
            if verb == "POST":
                func = requests.post
            elif verb == "GET":
                func = requests.get
            else:
                raise Exception(f"Unsupported method: {verb} for {endpoint}")
            response = func(
                f"{self._api_url}{endpoint}",
                json=payload,
                headers=headers,
            )
            if response.status_code == 200:
                return response.json()
            else:
                raise BreatheException(f"Error {response.status_code}: {response.text}")
        except Exception as e:
            raise BreatheException(f"Error {e}")

    def _make_model_api_call(self, endpoint: str, payload: dict | None) -> dict:
        full_endpoint = f"/model-services/pbm/{endpoint}"
        return self._make_api_call("POST", full_endpoint, payload)

    def get_service_version(self) -> Version:
        """Get the version number of the API

        Returns:
            _type_: _description_
        """
        response = self._make_model_api_call("get_version", payload={})

        version_str = response["versionStr"]
        return Version(version_str)

    def get_batteries(self) -> list[str]:
        """
        Gets a list of all batteries in your library.

        Returns:
            (list[str]): list of the available battery models.  Use these strings in subsequent functions for the "base_battery" argument.
        """
        _logger.info("Getting batteries...")
        batteries = self._make_api_call("GET", "/cells/", None)

        # unpack the ID and model name
        return [b["pbm_model_name"] for b in batteries]

    def get_battery_format(self, base_battery: str) -> dict:
        """
        Gets the battery format for a base battery. These are the form factor and cell casing parameters available for change in the api.

        Args:
            base_battery (str): The base battery model for the battery format.

        Returns:
            (dict): containing the battery format.
        """
        _logger.info(f"Getting battery format for {base_battery}...")
        return self._make_model_api_call(
            "get_battery_format", {"base_battery": base_battery}
        )

    def get_updated_format(self, base_battery: str, **kwargs) -> dict:
        """
        Gets the battery format for a base battery and updates it with the additional keywords. These are the form factor and cell casing parameters available for change in the api.
        Args:
            base_battery (str): The base battery model for the battery format.

        Returns:
            (dict): containing the battery format.
        """
        _logger.info(f"Getting battery format for {base_battery}...")
        format = self._make_model_api_call(
            "get_battery_format", {"base_battery": base_battery}
        )
        for key in kwargs.keys():
            if key in format or key == "name":
                format[key] = kwargs[key]
            else:
                raise BreatheException(f"Key {key} not found in the battery format.")
        return format

    def get_active_materials(self):
        """
        Gets a list of all acive materials.

        Returns:
            (pandas.Dataframe): containing the battery active materials.
        """
        _logger.info("Getting active materials...")
        am = self._make_model_api_call("get_active_materials", {})
        return pd.DataFrame(am)

    def get_design_parameters(self, base_battery: str):
        """
        Gets the design parameters for a base battery.

        Args:
            base_battery (str): The base battery model for the design parameters.

        Returns:
            (dict): containing the design parameters.
        """
        _logger.info(f"Getting design parameters for {base_battery}...")
        return self._make_model_api_call(
            "get_design_parameters", {"base_battery": base_battery}
        )

    def download_designs(
        self,
        base_battery: str,
        designs: list[dict] = [],
        formats: list[dict] = [],
        output_tag: str = "",
        folder: str = ".",
    ):
        """Send the design parameters for a given battery to the API, and get the resulting design, to be used with the Simulink blocks.

        Args:
            base_battery (str): The base battery to use to generate the design
            designs (list[str], optional): The designs for the simulation.
            formats (list[str], optional): The formats for the simulation.
            output_tag (str): A tag to be included in the metadata description.
            folder (str): The folder where design files will be saved. Defaults to current directory.
        Raises:
            BreatheException: _description_
        """
        import glob

        map_h2m, map_m2h = make_design_names_map(designs)
        designs_mapped = map_design_names__human_to_machine(designs, map_h2m)
        design_dict_mapped = self._make_model_api_call(
            "download_design",
            {
                "base_battery": base_battery,
                "designs": designs_mapped,
                "formats": formats,
            },
        )

        # undo design name mapping
        design_dict = {}
        for k, v in design_dict_mapped.items():
            key_human = map_m2h.get(k, k)
            design_dict[key_human] = v
            design_dict[key_human]["Meta"]["cellName"] = key_human

        # Ensure folder exists
        os.makedirs(folder, exist_ok=True)

        # Find any existing files in the provided directory (to avoid trying to save with the same name and overwriting)
        existing_files = glob.glob(os.path.join(folder, "BMJN_*.json"))
        existing_numbers = set()
        pattern = re.compile(r"BMJN_(\d+)\.json$")
        for file_path in existing_files:
            filename = os.path.basename(file_path)
            match = pattern.match(filename)
            if match:
                try:
                    number = int(match.group(1))
                    existing_numbers.add(number)
                except ValueError:
                    continue

        # Find the next available number
        next_number = 0
        while next_number in existing_numbers:
            next_number += 1

        design_names = design_dict.keys()

        # Prepare metadata for description
        metadata_items = [
            f"breathe_design version: {__version__}",
            f"base_battery: {base_battery}",
        ]
        if output_tag:
            metadata_items.append(f"output_tag: {output_tag}")
        if formats:
            format_names = [fmt.get("name", "unnamed_format") for fmt in formats]
            metadata_items.append(f"formats: {', '.join(format_names)}")

        for i, design_name in enumerate(design_names):
            # Use BMJN_XXXX naming convention
            filename = f"BMJN_{next_number + i:04d}.json"
            filepath = os.path.join(folder, filename)

            # Add metadata to the design
            design_data = design_dict[design_name].copy()
            if "Meta" not in design_data:
                design_data["Meta"] = {}
            if "description" not in design_data["Meta"]:
                design_data["Meta"]["description"] = []

            # Add the metadata items to description
            for item in metadata_items:
                if item not in design_data["Meta"]["description"]:
                    design_data["Meta"]["description"].append(item)

            with open(filepath, "w") as f:
                json.dump(design_data, f)

    def get_ocv(self, base_battery: str) -> pd.DataFrame:
        """
        Gets the OCV summary table for a base battery.

        Args:
            base_battery (str): The base battery model for the design parameters.

        Returns:
            (pandas.DataFrame): containing the OCV summary table.
        """
        _logger.info(f"Getting OCV for {base_battery}...")
        ocv_data = self._make_model_api_call("get_ocv", {"base_battery": base_battery})

        ocv_df = pd.DataFrame(ocv_data)
        return ocv_df

    def get_aged_ocv(
        self,
        base_battery: str,
        LAMPE: float = 0.0,
        LAMNE: float = 0.0,
        LLI: float = 0.0,
    ) -> pd.DataFrame:
        """
        Gets the OCV summary table for a base battery.

        Args:
            base_battery (str): The base battery model for the design parameters.
            LAMPE (float, optional): The electrolyte loss at positive electrode (PE) side.
            LAMNE (float, optional): The electrolyte loss at negative electrode (NE) side.
            LLI (float, optional): The loss of lithium inventory.

        Returns:
            (pandas.DataFrame): containing the OCV summary table.
        """
        _logger.info(f"Getting aged OCV for {base_battery}...")
        ocv_data = self._make_model_api_call(
            "get_aged_ocv",
            {"base_battery": base_battery, "LAMPE": LAMPE, "LAMNE": LAMNE, "LLI": LLI},
        )

        ocv_df = pd.DataFrame(ocv_data)
        return ocv_df

    def _format_kpis(self, results: dict) -> pd.DataFrame:
        kpis = results["KPIs"]
        kpi_names = [_transform_key(key) for key in results["KPInames"]]
        results_df = pd.DataFrame(kpis)
        # Add KPI names as a column to the DataFrame
        results_df["KPI"] = kpi_names
        results_df = results_df[
            ["KPI"] + [col for col in results_df.columns if col != "KPI"]
        ]
        return results_df.set_index("KPI")

    def get_eqm_kpis(
        self, base_battery: str, designs: list[dict] = [], formats: list[dict] = []
    ) -> "SingleSimulationResults":
        """
        Gets the equilibrium kpis for a given base battery, designs, and initial conditions.

        Args:
            base_battery (str): The base battery model for the simulation.
            designs (list[dict], optional): The designs for the simulation.
            formats (list[dict], optional): The battery formats for the simulation.

        Returns:
            (SingleSimulationResults): A results handler object containing the equilibrium KPIs.
                Use the plot_sensitivities() method to generate sensitivity plots.
        """
        _logger.info(f"Getting equilibrium kpis for {base_battery}...")
        map_h2m, map_m2h = make_design_names_map(designs)
        designs_mapped = map_design_names__human_to_machine(designs, map_h2m)
        results = self._make_model_api_call(
            "run_sim",
            {
                "base_battery": base_battery,
                "designs": designs_mapped,
                "formats": formats,
            },
        )

        # unmap the KPIs
        map_fields_in_place(results, "KPIs", map_m2h)

        # Add input parameters to the result for easy access
        results["input_parameters"] = {
            "base_battery": base_battery,
            "designs": designs,
            "formats": formats,
        }

        return SingleSimulationResults(results)

    def run_sim(
        self,
        base_battery: str,
        cycler: dict,
        designs: list[dict] = [],
        formats: list[dict] = [],
        initialSoC: Union[int, float, List[Union[int, float]]] = 0.5,
        initialTemperature_degC: Union[int, float, List[Union[int, float]]] = 25.0,
        ambientTemperature_degC: Union[int, float, List[Union[int, float]]] = 25.0,
    ) -> Union["SingleSimulationResults", "BatchSimulationResults"]:
        """
        Runs a simulation for a given base battery, cycler, designs, and initial conditions.

        Args:
            base_battery (str): The base battery model for the simulation.
            cycler (dict): The cycler parameters for the simulation.
            designs (list[dict], optional): The design parameters for the simulation.
            formats (list[dict], optional): The battery formats for the simulation.
            initialSoC (int | float | list[int | float], optional): The initial state of charge for the simulation.
                If a single value is provided, it will be used for all simulations.
                If a list of values is provided, multiple simulations will be run with each value.
            initialTemperature_degC (int | float | list[int | float], optional): The initial temperature for the simulation in degC.
                If a single value is provided, it will be used for all simulations.
                If a list of values is provided, multiple simulations will be run with each value.
            ambientTemperature_degC (int | float | list[int | float], optional): The ambient temperature for the simulation in degC.
                If a single value is provided, it will be used for all simulations.
                If a list of values is provided, multiple simulations will be run with each value.

        Returns:
            (SimulationResults): A results handler object containing the simulation results.
                This automatically handles both single and batch simulations.
        """
        # Validate and format design parameters
        validated_designs = parameter_validator.validate_and_format_designs(designs)

        # Validate and format format parameters
        validated_formats = parameter_validator.validate_and_format_formats(formats)

        # map the designs to machine names
        map_h2m, map_m2h = make_design_names_map(validated_designs)
        designs_mapped = map_design_names__human_to_machine(validated_designs, map_h2m)

        if (
            isinstance(initialSoC, (int, float))
            and isinstance(initialTemperature_degC, (int, float))
            and isinstance(ambientTemperature_degC, (int, float))
        ):
            # Single value for initialSoC and initialTemperature_degC
            payload = {
                "base_battery": base_battery,
                "cycler": cycler,
                "designs": designs_mapped,
                "formats": validated_formats,
                "initial_conditions": {
                    "socCell_0": initialSoC,
                    "initialTemperature_degC": initialTemperature_degC,
                    "ambientTemperature_degC": ambientTemperature_degC,
                },
            }

            result = self._make_model_api_call("run_sim", payload)

            # Convert "none" strings to Python None in dynamic data
            if "dynamicData" in result:
                result["dynamicData"] = convert_none_strings_to_none(
                    result["dynamicData"]
                )

            map_fields_in_place(result, "KPIs", map_m2h)
            map_fields_in_place(result, "dynamicData", map_m2h)

            # Add input parameters to the result for easy access
            result["input_parameters"] = {
                "base_battery": base_battery,
                "cycler": cycler,
                "designs": designs,
                "initialSoC": initialSoC,
                "initialTemperature_degC": initialTemperature_degC,
                "ambientTemperature_degC": ambientTemperature_degC,
            }

            return SingleSimulationResults(result)
        else:
            # make sure we are logged in first.  We need to do this because we are about to spawn multiple
            # threads, and if we do that when we are logged out, each thread will independencly try to log back in.
            self.ensure_logged_in()

            initialSoCs = (
                [initialSoC] if not isinstance(initialSoC, list) else initialSoC
            )
            initialTemperature_degC = (
                [initialTemperature_degC]
                if not isinstance(initialTemperature_degC, list)
                else initialTemperature_degC
            )
            ambientTemperature_degC = (
                [ambientTemperature_degC]
                if not isinstance(ambientTemperature_degC, list)
                else ambientTemperature_degC
            )
            max_workers = min(4, len(initialSoCs) * len(initialTemperature_degC))
            results = []
            all_payloads = [
                {
                    "base_battery": base_battery,
                    "cycler": cycler,
                    "designs": designs_mapped,
                    "formats": validated_formats,
                    "initial_conditions": {
                        "socCell_0": isc,
                        "initialTemperature_degC": itmp,
                        "ambientTemperature_degC": atmp,
                    },
                }
                for isc in initialSoCs
                for itmp in initialTemperature_degC
                for atmp in ambientTemperature_degC
            ]

            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                for i, result in enumerate(
                    executor.map(
                        self._make_model_api_call,
                        ["run_sim"] * len(all_payloads),
                        all_payloads,
                    )
                ):
                    if result is None:
                        raise BreatheException(
                            f"Simulation run for conditions {all_payloads[i]} failed!"
                        )

                    # Convert "none" strings to Python None in dynamic data
                    if "dynamicData" in result:
                        result["dynamicData"] = convert_none_strings_to_none(
                            result["dynamicData"]
                        )

                    map_fields_in_place(result, "KPIs", map_m2h)
                    map_fields_in_place(result, "dynamicData", map_m2h)

                    # Add input parameters to each result for easy access
                    result["input_parameters"] = {
                        "base_battery": base_battery,
                        "cycler": cycler,
                        "designs": designs,
                        "initialSoC": all_payloads[i]["initial_conditions"][
                            "socCell_0"
                        ],
                        "initialTemperature_degC": all_payloads[i][
                            "initial_conditions"
                        ]["initialTemperature_degC"],
                        "ambientTemperature_degC": all_payloads[i][
                            "initial_conditions"
                        ]["ambientTemperature_degC"],
                    }

                    results.append(result)

            return BatchSimulationResults(results)


api_interface = BreatheDesignModel()



#!/usr/bin/env python3
"""
Script to create a public release by:
1. Cloning the GitHub repository into public_repo folder
2. Pulling latest from main in the cloned repo
3. Creating a release branch with version and commit hash
4. Copying files from this repo to the cloned repo
5. Committing changes and creating a pull request
"""

import subprocess
import sys
import argparse
from pathlib import Path
import re
import requests
import os
import shutil


def run_command(cmd, cwd=None, capture_output=True):
    """Run a shell command and return the result"""
    try:
        if isinstance(cmd, str):
            cmd = cmd.split()

        result = subprocess.run(
            cmd, cwd=cwd, capture_output=capture_output, text=True, check=True
        )
        return result.stdout.strip() if capture_output else ""
    except subprocess.CalledProcessError as e:
        print(f"✗ Command failed: {' '.join(cmd)}")
        print(f"Error: {e.stderr if e.stderr else e}")
        raise


def get_version_from_init():
    """Extract version from breathe_design/__init__.py"""
    init_file = Path(__file__).parent.parent / "breathe_design" / "__init__.py"

    if not init_file.exists():
        raise FileNotFoundError(f"Could not find __init__.py at {init_file}")

    with open(init_file, "r") as f:
        content = f.read()

    # Look for __version__ = "x.y.z" pattern
    version_match = re.search(r'__version__\s*=\s*["\']([^"\']+)["\']', content)
    if not version_match:
        raise ValueError("Could not find __version__ in __init__.py")

    return version_match.group(1)


def get_current_commit_hash():
    """Get the current commit hash of this repo"""
    return run_command("git rev-parse --short HEAD")


def get_current_branch():
    """Get the current branch name of this repo"""
    return run_command("git rev-parse --abbrev-ref HEAD")


def clone_github_repo(github_url, target_path):
    """Clone the GitHub repository into the target path"""
    print(f"Cloning GitHub repository: {github_url}")

    # Remove existing directory if it exists
    if target_path.exists():
        print(f"Removing existing directory: {target_path}")
        shutil.rmtree(target_path)

    # Clone the repository
    run_command(f"git clone {github_url} {target_path}")
    print(f"✓ Successfully cloned repository to {target_path}")


def pull_latest_main(repo_path):
    """Pull the latest changes from main branch in the cloned repo"""
    print("Pulling latest changes from main in cloned repo...")

    # Change to repo directory and pull latest main
    run_command("git checkout main", cwd=repo_path)
    run_command("git pull origin main", cwd=repo_path)

    print("✓ Successfully pulled latest main")


def create_release_branch(repo_path, version, commit_hash, source_branch):
    """Create a new release branch in the cloned repo"""
    branch_name = f"release_from_branch_{source_branch}_{version}_{commit_hash}"

    print(f"Creating release branch: {branch_name}")

    # Create and checkout new branch
    run_command(f"git checkout -b {branch_name}", cwd=repo_path)

    print(f"✓ Created and switched to branch: {branch_name}")
    return branch_name


def run_copy_script(repo_path):
    """Run the copy script to sync files to the cloned repo"""
    print("Running copy script to sync files...")

    # Run the copy script with the repo path as destination
    copy_script = Path(__file__).parent / "copy_to_remote_repo.py"
    run_command(f"python {copy_script} {repo_path}", capture_output=False)

    print("✓ Files copied successfully")


def commit_changes(repo_path, version, commit_hash):
    """Add all changes and commit them in the cloned repo"""
    print("Committing changes in cloned repo...")

    # Add all changes
    run_command(["git", "add", "."], cwd=repo_path)

    # Check if there are any changes to commit
    try:
        status = run_command("git status --porcelain", cwd=repo_path)
        if not status.strip():
            print("⚠ No changes to commit")
            return False
    except subprocess.CalledProcessError:
        pass

    # Commit changes
    commit_message = f"Release {version} from commit {commit_hash}"
    run_command(["git", "commit", "-m", commit_message], cwd=repo_path)

    print(f"✓ Committed changes: {commit_message}")
    return True


def push_branch(repo_path, branch_name):
    """Push the release branch to origin"""
    print(f"Pushing branch {branch_name} to origin...")

    try:
        run_command(f"git push -u origin {branch_name}", cwd=repo_path)
        print(f"✓ Pushed branch {branch_name}")
        return True
    except subprocess.CalledProcessError as e:
        if "403" in str(e) or "not allowed to push" in str(e).lower():
            print("✗ Push failed: Permission denied")
            print(
                "This usually means the CI job doesn't have write access to the GitHub repository."
            )
            print("Solutions:")
            print("1. Create a Personal Access Token with 'repo' scope")
            print("2. Set GITHUB_TOKEN as a CI/CD variable")
            print(f"3. The branch '{branch_name}' was created locally but not pushed")
            return False
        else:
            raise


def get_github_token():
    """Get GitHub access token from environment or user input"""
    # Try to get token from environment variable
    token = os.getenv("GITHUB_TOKEN")
    if token:
        return token

    # Try to get token from git config (if set)
    try:
        token = run_command("git config --global github.token")
        if token:
            return token
    except subprocess.CalledProcessError:
        pass

    # Ask user for token
    print("GitHub access token required for creating pull requests.")
    print("You can:")
    print("1. Set GITHUB_TOKEN environment variable")
    print("2. Set git config: git config --global github.token YOUR_TOKEN")
    print("3. Get a token from: https://github.com/settings/tokens")
    print("   (Needs 'repo' scope)")
    print()

    token = input("Enter your GitHub access token (or press Enter to skip): ").strip()
    return token if token else None


def extract_github_repo_info(repo_path):
    """Extract GitHub repository info from git remote"""
    try:
        # Get the remote URL
        remote_url = run_command("git remote get-url origin", cwd=repo_path)

        # Parse GitHub URL to get owner/repo
        # Example: https://github.com/InfraBreathe/breathe_design.git
        if "github.com" in remote_url:
            # Remove .git suffix and extract path
            repo_path_str = remote_url.replace("https://github.com/", "").replace(
                ".git", ""
            )
            return "github.com", repo_path_str
        else:
            print(f"⚠ Non-GitHub.com URL detected: {remote_url}")
            return None, None

    except subprocess.CalledProcessError:
        print("✗ Could not get git remote URL")
        return None, None


def create_pull_request(repo_path, branch_name, version, commit_hash):
    """Create a pull request using GitHub REST API"""
    print("Creating pull request...")

    # Get GitHub repository info
    github_host, repo_path_str = extract_github_repo_info(repo_path)
    if not github_host or not repo_path_str:
        print("✗ Could not extract GitHub repository information")
        return False

    # Get access token
    token = get_github_token()
    if not token:
        print("⚠ No GitHub token provided. Please create pull request manually:")
        print("  1. Go to the GitHub repository")
        print(f"  2. Create a pull request from branch '{branch_name}' to 'main'")
        print(f"  3. Title: 'Release {version} from commit {commit_hash}'")
        print(f"  4. Description: Release of breathe_design version {version}")
        return False

    try:
        # Prepare pull request data
        title = f"Release {version} from commit {commit_hash}"
        description = f"""Release of breathe_design version {version} from source commit {commit_hash}.

This release includes:
- Updated documentation and examples
- Latest version of requirements.txt
- All necessary files for public distribution

Auto-generated by create_public_release.py"""

        # GitHub API endpoint
        api_url = f"https://api.github.com/repos/{repo_path_str}/pulls"

        # Request payload
        payload = {
            "title": title,
            "head": branch_name,
            "base": "main",
            "body": description,
        }

        # Request headers
        headers = {
            "Authorization": f"token {token}",
            "Accept": "application/vnd.github.v3+json",
        }

        print(f"Creating PR at: {api_url}")
        print(f"From: {branch_name} → main")

        # Make the API request
        response = requests.post(api_url, json=payload, headers=headers, timeout=30)

        if response.status_code == 201:
            pr_data = response.json()
            pr_url = pr_data.get("html_url", "")
            pr_number = pr_data.get("number", "")
            print("✓ Pull request created successfully!")
            print(f"  URL: {pr_url}")
            print(f"  PR #{pr_number}: {title}")
            return True
        else:
            print(f"✗ Failed to create pull request (HTTP {response.status_code})")
            try:
                error_info = response.json()
                if "message" in error_info:
                    print(f"  Error: {error_info['message']}")
                if "errors" in error_info:
                    for error in error_info["errors"]:
                        print(f"  Details: {error}")
            except ValueError:
                print(f"  Response: {response.text}")
            return False

    except requests.exceptions.RequestException as e:
        print(f"✗ Network error creating pull request: {e}")
        return False
    except Exception as e:
        print(f"✗ Error creating pull request: {e}")
        return False


def main():
    parser = argparse.ArgumentParser(
        description="Create a public release in the GitHub repository"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Show what would be done without actually doing it",
    )
    parser.add_argument(
        "--skip-pr",
        action="store_true",
        help="Skip creating the pull request",
    )

    args = parser.parse_args()

    # Get paths
    current_repo = Path(
        __file__
    ).parent.parent  # Go up one level from public_release_scripts
    github_repo_path = current_repo / "public_repo"

    # GitHub repository URL
    github_url = os.getenv(
        "GITHUB_REPO", "https://github.com/InfraBreathe/breathe_design.git"
    )

    try:
        # Get version, commit info, and source branch
        version = get_version_from_init()
        commit_hash = get_current_commit_hash()
        source_branch = get_current_branch()

        print(f"Source repo: {current_repo}")
        print(f"GitHub repo: {github_url}")
        print(f"Target path: {github_repo_path}")
        print(f"Version: {version}")
        print(f"Commit hash: {commit_hash}")
        print(f"Source branch: {source_branch}")
        print()

        if args.dry_run:
            print("DRY RUN - What would happen:")
            print("1. Clone GitHub repository")
            print("2. Pull latest main in cloned repo")
            print(
                f"3. Create branch: release_from_branch_{source_branch}_{version}_{commit_hash}"
            )
            print("4. Run copy script to sync files")
            print("5. Commit all changes")
            print("6. Push branch to origin")
            if not args.skip_pr:
                print("7. Create pull request")
            return

        # Step 1: Clone GitHub repository
        clone_github_repo(github_url, github_repo_path)
        print()

        # Step 2: Pull latest main
        pull_latest_main(github_repo_path)
        print()

        # Step 3: Create release branch
        branch_name = create_release_branch(
            github_repo_path, version, commit_hash, source_branch
        )
        print()

        # Step 4: Run copy script
        run_copy_script(github_repo_path)
        print()

        # Step 5: Commit changes
        has_changes = commit_changes(github_repo_path, version, commit_hash)
        if not has_changes:
            print("No changes to release. Exiting.")
            return
        print()

        # Step 6: Push branch
        push_success = push_branch(github_repo_path, branch_name)
        if not push_success:
            print(
                "⚠ Branch push failed. Release branch was created locally but not pushed to remote."
            )
            print(
                "You may need to push manually or fix the authentication configuration."
            )
        print()

        # Step 7: Create pull request (if not skipped and push was successful)
        if not args.skip_pr and push_success:
            pr_success = create_pull_request(
                github_repo_path, branch_name, version, commit_hash
            )
            if not pr_success:
                print(
                    "⚠ Pull request creation failed, but release branch was created successfully"
                )
        elif not args.skip_pr and not push_success:
            print("⚠ Skipping pull request creation due to push failure")

        # Summary
        print("=" * 60)
        print("RELEASE SUMMARY")
        print("=" * 60)
        print(f"Version: {version}")
        print(f"Source commit: {commit_hash}")
        print(f"Release branch: {branch_name}")
        print(
            f"Branch pushed: {'✓' if push_success else '✗ (failed - check permissions)'}"
        )
        print(f"GitHub repo: {github_url}")
        print(f"Local path: {github_repo_path}")

    except Exception as e:
        print(f"✗ Error: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd


def extract_dynamic_kpis(dynamic_kpi, data):
    """
    Extract KPIs from the given data and return them in a pandas DataFrame.

    Parameters:
        data (dict): Dictionary containing time series data for each design. Expected format: {design_name: {'Time': [...], 'voltModel': [...], 'currentModel': [...]}, ...}

    Returns:
        (pandas.DataFrame): DataFrame with columns ['Design', 'Resistance', 'Capacity'] containing the KPIs for each design
    """
    if dynamic_kpi not in ["DCIR", "RateCap"]:
        return None
    # Initialize lists to store results
    designs = []
    resistances = []
    capacities = []

    # Process each design
    for design_name in data.keys():
        t = np.asarray(data[design_name]["Time"])
        v = np.asarray(data[design_name]["voltModel"])
        i = np.asarray(data[design_name]["currModel"])

        # Find the last point before current goes negative (discharge starts)
        discharge_start_idx = np.where(i < 0)[0]
        if len(discharge_start_idx) > 0:
            # Get the voltage just before discharge starts
            v_before_discharge = (
                v[discharge_start_idx[0] - 1] if discharge_start_idx[0] > 0 else v[0]
            )
        else:
            # If no discharge found, use first voltage
            v_before_discharge = v[0]

        i_map = i < 0
        t = t[i_map]
        v = v[i_map]
        i = i[i_map]
        # Calculate resistance and capacity
        resistance = np.abs((v_before_discharge - v[-1]) / np.mean(i))
        capacity = np.abs(np.trapezoid(i, t) / 3600)

        # Store results
        designs.append(design_name)
        resistances.append(resistance)
        capacities.append(capacity)

    # Create DataFrame
    if dynamic_kpi == "DCIR":
        results_df = pd.DataFrame({"Design": designs, "Resistance (Ω)": resistances})
    elif dynamic_kpi == "RateCap":
        results_df = pd.DataFrame({"Design": designs, "Capacity (Ah)": capacities})

    return results_df



import pytest
import copy
from breathe_design.parameter_validator import ParameterValidator, parameter_validator


class TestParameterValidator:
    """Test cases for the unified ParameterValidator class."""

    def test_validator_initialization(self):
        """Test that the validator initializes correctly."""
        validator = ParameterValidator()
        assert validator._parameter_bounds is not None
        assert isinstance(validator._parameter_bounds, dict)
        assert len(validator._parameter_bounds) > 0

    def test_singleton_instance(self):
        """Test that the singleton instance is properly created."""
        assert parameter_validator is not None
        assert isinstance(parameter_validator, ParameterValidator)

    def test_get_parameter_bounds_all(self):
        """Test getting all parameter bounds."""
        bounds = parameter_validator.get_parameter_bounds()
        assert isinstance(bounds, dict)
        assert len(bounds) > 0

        # Check some known parameters
        assert "NPratio" in bounds
        assert "housingThickness_mm" in bounds
        assert "length_mm" in bounds

        # Check bounds format
        assert bounds["NPratio"] == [0.7, 3]
        assert bounds["housingThickness_mm"] == [0.1, 10.0]

    def test_get_parameter_bounds_specific(self):
        """Test getting bounds for a specific parameter."""
        bounds = parameter_validator.get_parameter_bounds("NPratio")
        assert bounds == {"NPratio": [0.7, 3]}

    def test_get_parameter_bounds_unknown(self):
        """Test getting bounds for an unknown parameter raises error."""
        with pytest.raises(ValueError, match="Unknown parameter"):
            parameter_validator.get_parameter_bounds("unknown_parameter")

    def test_validate_design_valid(self):
        """Test validation of a valid design."""
        design = {
            "designName": "Test Design",
            "NPratio": 1.5,
            "Vmin_V": 3.0,
            "Vmax_V": 4.2,
            "cathodePorosity": 0.4,
            "anodePorosity": 0.3,
        }

        result = parameter_validator.validate_design(design)
        assert result["designName"] == "Test Design"
        assert result["NPratio"] == 1.5
        assert result["Vmin_V"] == 3.0

    def test_validate_design_out_of_bounds_high(self):
        """Test validation fails for parameter above maximum."""
        design = {
            "designName": "Invalid Design",
            "NPratio": 5.0,  # Above maximum of 3.0
            "Vmin_V": 3.0,
        }

        with pytest.raises(ValueError, match="above maximum allowed value"):
            parameter_validator.validate_design(design)

    def test_validate_design_out_of_bounds_low(self):
        """Test validation fails for parameter below minimum."""
        design = {
            "designName": "Invalid Design",
            "NPratio": 0.5,  # Below minimum of 0.7
            "Vmin_V": 3.0,
        }

        with pytest.raises(ValueError, match="below minimum allowed value"):
            parameter_validator.validate_design(design)

    def test_validate_design_none_value(self):
        """Test validation fails for None parameter value."""
        design = {"designName": "Invalid Design", "NPratio": None, "Vmin_V": 3.0}

        with pytest.raises(ValueError, match="cannot be None"):
            parameter_validator.validate_design(design)

    def test_validate_design_non_numeric(self):
        """Test validation fails for non-numeric parameter."""
        design = {
            "designName": "Invalid Design",
            "NPratio": "not_a_number",
            "Vmin_V": 3.0,
        }

        with pytest.raises(ValueError, match="must be numeric"):
            parameter_validator.validate_design(design)

    def test_validate_design_unknown_parameter_allowed(self):
        """Test that unknown parameters are allowed (for future extensibility)."""
        design = {
            "designName": "Test Design",
            "NPratio": 1.5,
            "unknown_future_parameter": 42.0,
        }

        # Should not raise an error
        result = parameter_validator.validate_design(design)
        assert result["designName"] == "Test Design"
        assert result["NPratio"] == 1.5
        assert result["unknown_future_parameter"] == 42.0

    def test_validate_designs_list(self):
        """Test validation of a list of designs."""
        designs = [
            {"designName": "Design 1", "NPratio": 1.0, "Vmin_V": 3.0},
            {"designName": "Design 2", "NPratio": 2.0, "Vmin_V": 2.8},
        ]

        result = parameter_validator.validate_designs(designs)
        assert len(result) == 2
        assert result[0]["designName"] == "Design 1"
        assert result[1]["designName"] == "Design 2"

    def test_validate_designs_list_with_error(self):
        """Test validation of design list fails on invalid design."""
        designs = [
            {"designName": "Valid Design", "NPratio": 1.0, "Vmin_V": 3.0},
            {
                "designName": "Invalid Design",
                "NPratio": 5.0,
                "Vmin_V": 2.8,
            },  # NPratio too high
        ]

        with pytest.raises(ValueError, match="Design 1 validation failed"):
            parameter_validator.validate_designs(designs)

    def test_validate_designs_not_list(self):
        """Test that validation fails if designs is not a list."""
        with pytest.raises(ValueError, match="must be a list"):
            parameter_validator.validate_designs({"not": "a_list"})

    def test_validate_format_valid_cuboid(self):
        """Test validation of a valid cuboid format."""
        format_dict = {
            "shape": "cuboid",
            "name": "Test Cuboid",
            "material": "aluminum",
            "housingThickness_mm": 1.0,
            "length_mm": 100.0,
            "width_mm": 50.0,
            "thickness_mm": 10.0,
            "assemblyType": "stacked",
        }

        result = parameter_validator.validate_format(format_dict)
        assert result["shape"] == "cuboid"
        assert result["material"] == "aluminum"
        assert result["housingThickness_mm"] == 1.0

    def test_validate_format_valid_cylinder(self):
        """Test validation of a valid cylinder format."""
        format_dict = {
            "shape": "cylinder",
            "name": "Test Cylinder",
            "material": "steel",
            "housingThickness_mm": 0.5,
            "diameter_mm": 18.0,
            "height_mm": 65.0,
            "innerDiameter_mm": 17.0,
        }

        result = parameter_validator.validate_format(format_dict)
        assert result["shape"] == "cylinder"
        assert result["material"] == "steel"

    def test_validate_format_missing_shape(self):
        """Test validation fails when shape is missing."""
        format_dict = {
            "name": "No Shape",
            "material": "aluminum",
            "housingThickness_mm": 1.0,
        }

        with pytest.raises(ValueError, match="must specify a 'shape' field"):
            parameter_validator.validate_format(format_dict)

    def test_validate_format_missing_material(self):
        """Test validation fails when material is missing."""
        format_dict = {
            "shape": "cuboid",
            "name": "No Material",
            "housingThickness_mm": 1.0,
        }

        with pytest.raises(ValueError, match="must specify a 'material' field"):
            parameter_validator.validate_format(format_dict)

    def test_validate_format_invalid_material(self):
        """Test validation fails for invalid material."""
        format_dict = {
            "shape": "cuboid",
            "name": "Invalid Material",
            "material": "plastic",  # Not in valid materials
            "housingThickness_mm": 1.0,
        }

        with pytest.raises(ValueError, match="Invalid material"):
            parameter_validator.validate_format(format_dict)

    def test_validate_format_invalid_assembly_type(self):
        """Test validation fails for invalid assembly type."""
        format_dict = {
            "shape": "cuboid",
            "name": "Invalid Assembly",
            "material": "aluminum",
            "housingThickness_mm": 1.0,
            "assemblyType": "invalid_type",
        }

        with pytest.raises(ValueError, match="Invalid assembly type"):
            parameter_validator.validate_format(format_dict)

    def test_validate_format_cylinder_constraints(self):
        """Test cylinder-specific constraint validation."""
        format_dict = {
            "shape": "cylinder",
            "name": "Invalid Cylinder",
            "material": "steel",
            "housingThickness_mm": 0.5,
            "diameter_mm": 18.0,
            "height_mm": 65.0,
            "innerDiameter_mm": 18.0,  # Equal to outer diameter - should fail
        }

        with pytest.raises(ValueError, match="must be less than outer diameter"):
            parameter_validator.validate_format(format_dict)

    def test_validate_format_parameter_out_of_bounds(self):
        """Test format validation fails for out-of-bounds parameter."""
        format_dict = {
            "shape": "cuboid",
            "name": "Too Thick",
            "material": "aluminum",
            "housingThickness_mm": 15.0,  # Above maximum of 10.0
            "length_mm": 100.0,
        }

        with pytest.raises(ValueError, match="above maximum allowed value"):
            parameter_validator.validate_format(format_dict)

    def test_validate_formats_list(self):
        """Test validation of a list of formats."""
        formats = [
            {
                "shape": "cuboid",
                "name": "Format 1",
                "material": "aluminum",
                "housingThickness_mm": 1.0,
                "length_mm": 100.0,
            },
            {
                "shape": "cylinder",
                "name": "Format 2",
                "material": "steel",
                "housingThickness_mm": 0.5,
                "diameter_mm": 18.0,
                "height_mm": 65.0,
                "innerDiameter_mm": 17.0,
            },
        ]

        result = parameter_validator.validate_formats(formats)
        assert len(result) == 2
        assert result[0]["shape"] == "cuboid"
        assert result[1]["shape"] == "cylinder"

    def test_format_parameters(self):
        """Test parameter formatting (type conversion)."""
        params = {
            "designName": "Test",
            "NPratio": 1,  # Integer
            "housingThickness_mm": 2,  # Integer
            "material": "aluminum",  # String - should not be converted
        }

        result = parameter_validator.format_parameters(params)
        assert result["designName"] == "Test"  # String unchanged
        assert result["NPratio"] == 1.0  # Converted to float
        assert result["housingThickness_mm"] == 2.0  # Converted to float
        assert result["material"] == "aluminum"  # String unchanged

    def test_validate_and_format_designs(self):
        """Test combined validation and formatting of designs."""
        designs = [
            {"designName": "Test", "NPratio": 1, "Vmin_V": 3}  # Integers
        ]

        result = parameter_validator.validate_and_format_designs(designs)
        assert len(result) == 1
        assert isinstance(result[0]["NPratio"], float)
        assert isinstance(result[0]["Vmin_V"], float)

    def test_validate_and_format_formats(self):
        """Test combined validation and formatting of formats."""
        formats = [
            {
                "shape": "cuboid",
                "material": "aluminum",
                "housingThickness_mm": 1,  # Integer
                "length_mm": 100,  # Integer
            }
        ]

        result = parameter_validator.validate_and_format_formats(formats)
        assert len(result) == 1
        assert isinstance(result[0]["housingThickness_mm"], float)
        assert isinstance(result[0]["length_mm"], float)

    def test_validate_design_not_dict(self):
        """Test validation fails if design is not a dictionary."""
        with pytest.raises(ValueError, match="must be a dictionary"):
            parameter_validator.validate_design("not_a_dict")

    def test_validate_format_not_dict(self):
        """Test validation fails if format is not a dictionary."""
        with pytest.raises(ValueError, match="must be a dictionary"):
            parameter_validator.validate_format("not_a_dict")

    def test_parameter_bounds_immutability(self):
        """Test that returned bounds are deep copies and don't affect internal state."""
        bounds1 = parameter_validator.get_parameter_bounds()
        bounds2 = parameter_validator.get_parameter_bounds()

        # Modify one copy
        bounds1["NPratio"] = [999, 999]

        # Other copy should be unchanged
        assert bounds2["NPratio"] == [0.7, 3]

        # Original validator should be unchanged
        bounds3 = parameter_validator.get_parameter_bounds()
        assert bounds3["NPratio"] == [0.7, 3]

    def test_design_immutability(self):
        """Test that validation doesn't modify the original design."""
        original_design = {"designName": "Test", "NPratio": 1.5}
        design_copy = copy.deepcopy(original_design)

        parameter_validator.validate_design(design_copy)

        # Original should be unchanged
        assert original_design == design_copy

    def test_format_immutability(self):
        """Test that validation doesn't modify the original format."""
        original_format = {
            "shape": "cuboid",
            "material": "aluminum",
            "housingThickness_mm": 1.0,
        }
        format_copy = copy.deepcopy(original_format)

        parameter_validator.validate_format(format_copy)

        # Original should be unchanged
        assert original_format == format_copy



VALID_MATERIALS = {"steel", "aluminum", "pouch"}
VALID_ASSEMENT_TYPES = {"stacked", "wound", "zFolded"}


def validate_material(material):
    if material not in VALID_MATERIALS:
        raise ValueError(
            f"Invalid material: '{material}'. Must be one of {sorted(VALID_MATERIALS)}."
        )


def validate_assembly_type(assembly_type):
    if assembly_type not in VALID_ASSEMENT_TYPES:
        raise ValueError(
            f"Invalid assembly type: '{assembly_type}'. Must be one of {sorted(VALID_ASSEMENT_TYPES)}."
        )


def validate_format_parameter(param_name, param_value):
    """Validate a format parameter against its bounds."""
    # Import here to avoid circular imports
    from .parameter_validator import parameter_validator

    bounds = parameter_validator.get_parameter_bounds()

    if param_name not in bounds:
        return  # Skip validation for unknown parameters

    if param_value is None:
        raise ValueError(f"Format parameter '{param_name}' cannot be None")

    if not isinstance(param_value, (int, float)):
        raise ValueError(
            f"Format parameter '{param_name}' must be numeric, got {type(param_value)}"
        )

    min_val, max_val = bounds[param_name]

    if min_val is not None and param_value < min_val:
        raise ValueError(
            f"Format parameter '{param_name}' value {param_value} is below minimum allowed value {min_val}"
        )

    if max_val is not None and param_value > max_val:
        raise ValueError(
            f"Format parameter '{param_name}' value {param_value} is above maximum allowed value {max_val}"
        )


def define_cuboid_format(
    name, material, housingThickness_mm, length_mm, width_mm, thickness_mm, assemblyType
):
    """
    Defines a dictionary for a cuboid battery format.

    Parameters:
        name (str): Name of the battery format
        material (str): Material used
        housingThickness_mm (float): Housing thickness
        length_mm (float): Length of the battery
        width_mm (float): Width of the battery
        thickness_mm (float): Thickness of the battery
        assembly_type (str): Type of assembly

    Returns:
        dict: A dictionary representing the cuboid battery format
    """
    validate_material(material)
    validate_assembly_type(assemblyType)

    # Validate numeric parameters
    validate_format_parameter("housingThickness_mm", housingThickness_mm)
    validate_format_parameter("length_mm", length_mm)
    validate_format_parameter("width_mm", width_mm)
    validate_format_parameter("thickness_mm", thickness_mm)

    return {
        "shape": "cuboid",
        "name": name,
        "material": material,
        "housingThickness_mm": float(housingThickness_mm),
        "length_mm": float(length_mm),
        "width_mm": float(width_mm),
        "thickness_mm": float(thickness_mm),
        "assemblyType": assemblyType,
    }


def define_cylinder_format(
    name, material, housingThickness_mm, diameter_mm, height_mm, innerDiameter_mm
):
    """
    Defines a dictionary for a cylindrical battery format.

    Parameters:
        name (str): Name of the battery format
        material (str): Material used
        housingThickness_mm (float): Housing thickness
        diameter_mm (float): Outer diameter of the cylinder
        height_mm (float): Height of the cylinder
        innerDiameter_mm (float): Inner diameter of the cylinder

    Returns:
        dict: A dictionary representing the cylindrical battery format
    """
    validate_material(material)

    # Validate numeric parameters
    validate_format_parameter("housingThickness_mm", housingThickness_mm)
    validate_format_parameter("diameter_mm", diameter_mm)
    validate_format_parameter("height_mm", height_mm)
    validate_format_parameter("innerDiameter_mm", innerDiameter_mm)

    # Additional validation: inner diameter should be less than outer diameter
    if innerDiameter_mm >= diameter_mm:
        raise ValueError(
            f"Inner diameter ({innerDiameter_mm}) must be less than outer diameter ({diameter_mm})"
        )

    return {
        "shape": "cylinder",
        "name": name,
        "material": material,
        "housingThickness_mm": float(housingThickness_mm),
        "diameter_mm": float(diameter_mm),
        "height_mm": float(height_mm),
        "innerDiameter_mm": float(innerDiameter_mm),
    }



__version__ = "0.6.5"
__date__ = "16/09/2025"

from .api_interface import *  # NOQA
from .cycler import *  # NOQA
from .plots import *  # NOQA
from .dynamic_analysis import *  # NOQA
from .utilities import *  # NOQA
from .format import *  # NOQA
from .results import *  # NOQA



import time
import json
import logging
import requests
import webbrowser
import tempfile
import os
import datetime
from threading import Lock

_logger = logging.getLogger(__name__)

temp_dir = tempfile.mkdtemp(prefix="myapp_")


class _DeviceFlowAuth:
    """Don't create an instance of this class, use the [device_auth] instance in this module."""

    def __init__(self, poll_interval=5):
        """
        Manages authentication using OAuth 2.0 Device Flow.

        :param poll_interval: How often to poll for token (in seconds).
        """

        self._poll_interval = poll_interval
        self._device_code_data: dict[str, str] | None = None
        self._token_data: dict[str, str] | None = None
        self._lock = Lock()
        self._last_time_token_checked = datetime.datetime.min

        tmp_dir_name = "devtmp"
        tmp_token_file = "dfile"
        self._tmp_dir_path = os.path.join(tempfile.gettempdir(), tmp_dir_name)
        self._tmp_token_file_path = os.path.join(self._tmp_dir_path, tmp_token_file)

        self._try_load_token_from_tmp()

        if "API_URL" in os.environ:
            self._api_url = os.environ["API_URL"]
            _logger.info(f"API_URL set from environment to {self._api_url}")
        else:
            self._api_url = (
                "https://breathe-platform-backend-service-prod.azurewebsites.net/api/v1"
            )

    def _try_load_token_from_tmp(self):
        """Try to load the token from a temporary file

        Returns:
            _type_: _description_
        """
        if not os.path.exists(self._tmp_dir_path):
            return False

        if not os.path.exists(self._tmp_token_file_path):
            return False

        try:
            with open(self._tmp_token_file_path, "r") as f:
                token_data = json.load(f)
                self._token_data = token_data
                return True
        except:  # noqa E722
            return False

    def _save_token_to_tmp(self):
        """Saves the token to a temporary file."""
        if self._token_data is None:
            return

        if not os.path.exists(self._tmp_dir_path):
            os.makedirs(self._tmp_dir_path)

        with open(self._tmp_token_file_path, "w") as f:
            json.dump(self._token_data, f)

    def _delete_tmp_token_file(self):
        if os.path.exists(self._tmp_token_file_path):
            os.remove(self._tmp_token_file_path)

    def _make_http_request(
        self,
        method,
        endpoint,
        payload: dict | None = None,
        headers=None,
    ):
        """Helper function to make HTTP requests using http.client."""

        response = requests.post(
            f"{self._api_url}{endpoint}", json=payload, headers=headers
        )
        return response

    def _get_device_code(self):
        """Requests a device code for user authentication."""
        response = self._make_http_request("POST", "/device-auth/code")
        if response.status_code != 200:
            print(f"⚠️ Failed to log in! {response.status_code}")
            return False

        response_data = response.json()
        self._device_code_data = response_data
        self._poll_interval = response_data["interval"]

        print(
            f"\n🚀 Open this URL in a browser and verify the code shown is {response_data['user_code']}"
        )
        verification_uri = response_data["verification_uri_complete"]
        print(f"🔗 {verification_uri}")
        webbrowser.open_new_tab(verification_uri)
        return True

    def _poll_for_token(self):
        """Polls the token endpoint until the user authorizes the device."""
        print("\n⏳ Waiting for user authorization...")
        device_code = self._device_code_data["device_code"]
        timeout_time = datetime.datetime.now() + datetime.timedelta(seconds=60)
        while True:
            try:
                response = self._make_http_request(
                    "POST",
                    "/device-auth/poll",
                    payload={"device_code": device_code},
                )

                response_data = response.json()
                if response.status_code == 200:
                    self._token_data = response_data
                    self._save_token_to_tmp()
                    print("\n✅ Authentication successful!")
                    return True
                    break
                elif response_data.get("error") == "authorization_pending":
                    time.sleep(self._poll_interval)  # Wait before polling again
                elif response_data.get("error") == "slow_down":
                    self._poll_interval += 5  # Adjust polling interval if required
                    time.sleep(self._poll_interval)
                else:
                    raise Exception(
                        f"⚠️ Authentication failed: {response_data.get('error_description', 'Unknown error')}"
                    )
            except Exception as e:
                error_message = e.args[0]  # Get the error message string

                # Find the position of `{` (start of JSON) and extract only the JSON part
                json_start = error_message.find("{")
                if json_start != -1:
                    try:
                        error_json = json.loads(
                            error_message[json_start:]
                        )  # Parse JSON
                        error_code = error_json.get("error")
                        error_description = error_json.get("error_description")
                        if error_code == "authorization_pending":
                            time.sleep(self._poll_interval)  # Wait before polling again
                        elif error_code == "slow_down":
                            self._poll_interval += (
                                5  # Adjust polling interval if required
                            )
                            time.sleep(self._poll_interval)
                        else:
                            raise Exception(
                                f"⚠️ Authentication failed: [{error_code}] {error_description}"
                            )
                    except json.JSONDecodeError:
                        print(
                            "❌ Failed to parse JSON from exception message.  Please re-run the command."
                        )
                else:
                    _logger.error(
                        f"Unhandled Exception: {e}.  Please re-run the command."
                    )
                    break

            if datetime.datetime.now() > timeout_time:
                print("❌ Failed to authenticate!  Please re-run the command.")
                return False

    def _start_auth_flow(self):
        print("\n🔐 No active session. Please sign in to continue.")
        if not self._get_device_code():
            return False
        return self._poll_for_token()

    def _check_if_expired(self, leeway: datetime.timedelta = datetime.timedelta()):
        """Check the token in self._token_data and check if it is expired

        If [leeway] is set then this function will consider the token expired if it is going to
        expire within [leeway] amount of time.

        Raises:
            RuntimeError: _description_

        Returns:
            bool: True if it is expired, False otherwise
        """

        if self._token_data is None:
            raise RuntimeError(
                "Can't check if token is expired because there is no token!  Please login first."
            )

        access_token = self._token_data["access_token"]
        response_data = self._make_http_request(
            "POST",
            "/device-auth/remaining-time",
            headers={
                "Authorization": f"Bearer {access_token}",
            },
        )
        if response_data.status_code == 401:
            # unauthorised so we need to login again
            return True

        if response_data.status_code != 200:
            raise RuntimeError(
                f"Failed to check token expiration! {response_data.status_code}"
            )
        response_data = response_data.json()

        # check expiry time including any leeway
        expired = response_data["time_left_s"] <= leeway.total_seconds()

        return expired

    def _get_new_token_from_refresh_token(self):
        """Retrives a new token using the refresh token.

        If successful, this updates the self._token_data with the new token data.

        Raises:
            RuntimeError: _description_
            RuntimeError: _description_
        """
        if (self._token_data is None) or ("refresh_token" not in self._token_data):
            print("No refresh token available.  You must log in again")
            self._token_data = None
            return False

        access_token = self._token_data["access_token"]
        refresh_token = self._token_data["refresh_token"]
        response = self._make_http_request(
            "POST",
            "/device-auth/refresh",
            payload={"refresh_token": refresh_token},
            headers={
                "accept": "application/json",
                "Content-Type": "application/json",
                "Authorization": f"Bearer {access_token}",
            },
        )
        if response.status_code != 200:
            # failed to refresh
            return False

        self._token_data = response.json()
        self._save_token_to_tmp()
        return True

    def _refresh_or_start(self):
        """Run token refresh using refresh token.
        If that fails, restart the auth flow to require login again

        Raises:
            RuntimeError: _description_
        """
        if not self._get_new_token_from_refresh_token():
            # if refreshing failed, we need to log in again
            if not self._start_auth_flow():
                raise RuntimeError("Failed to get access token")

    def get_token(self):
        """Returns the current access token. If missing, prompts the user to sign in."""
        with self._lock:
            if not self._token_data:
                if not self._start_auth_flow():
                    raise RuntimeError("Failed to get access token")

                #  immediately refresh because if this was just after signup, the token may not have the correct permissions set yet
                self._refresh_or_start()
            else:
                # we have a token, but check the expiry and if it will expire within the next 5 mins try to refresh it with the refresh token.
                # However if we already checked within the last 5 seconds, don't check again.
                # If we already checked within that time, then whatever the outcome was then won't have changed (because we check for expiry within the next 5 mins).
                # (we do this rate check this to prevent the token check endpoint from being called rapidly e.g. if there are multiple threads getting tokens, and
                # triggering a 429 on the endpoint)
                if (
                    datetime.datetime.now() - self._last_time_token_checked
                    > datetime.timedelta(seconds=5)
                ):
                    self._last_time_token_checked = datetime.datetime.now()
                    if self._check_if_expired(leeway=datetime.timedelta(minutes=5)):
                        self._refresh_or_start()
                else:
                    _logger.debug(
                        "Skipping token expiry check because it was checked recently"
                    )

            return self._token_data["access_token"]

    def logout(self):
        """Log out from the Breathe API.
        This consists of logging out of the OAuth system and then deleting the local token storage.
        """
        try:
            response = self._make_http_request(
                "POST",
                "/device-auth/logout",
                payload={"id_token": self._token_data["id_token"]},
                headers={
                    "accept": "application/json",
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {self._token_data['access_token']}",
                },
            )
            if response.status_code != 200:
                msg = f"Error logging out!: {response.status_code}"
                if response.content:
                    msg += f": {response.content}"
                _logger.error(msg)
        except:  # noqa
            # if we fail to log out, no matter, still delete the token requiring us to get a new one
            pass
        with self._lock:
            self._token_data = None
            self._delete_tmp_token_file()


device_auth = _DeviceFlowAuth()



import plotly.graph_objects as go
import plotly.express as px
import numpy as np
from .utilities import _transform_key, normalize_results

# Breathe Hex Codes
hex_colors = ["#00c34e", "#9bda9e", "#ffffff", "#e0dcda", "#b8aea7"]


def plot_sensitivities(kpi_names, sensitivities):
    """
    Create a heatmap plot of sensitivities for different KPIs and parameters.

    This function generates a heatmap visualization of the sensitivity values
    for various Key Performance Indicators (KPIs) with respect to different parameters.
    The heatmap is normalized for each KPI to enhance visual comparison.

    Args:
        kpi_names (list): A list of strings representing the names of the KPIs.
        sensitivities (dict): A dictionary where keys are parameter names and values are lists of sensitivity values corresponding to each KPI.

    Returns:
        (plotly.Figure): A Plotly Figure object representing the sensitivity heatmap, which can be displayed or further modified.
    """
    kpi_names = [_transform_key(key) for key in kpi_names]
    param_names = [_transform_key(key) for key in sensitivities.keys()]
    sensitivity_values = list(sensitivities.values())

    heatmap_data = np.array(sensitivity_values)
    for k in range(len(kpi_names)):
        normalizer = np.max(np.abs(heatmap_data[:, k]))
        if normalizer > 0:
            heatmap_data[:, k] *= 1 / normalizer

    fig = go.Figure(
        data=go.Heatmap(
            z=heatmap_data,
            x=kpi_names,
            y=param_names,
            colorscale=hex_colors,
            reversescale=True,
            text=[[f"{val:.2f}" for val in row] for row in heatmap_data],
            texttemplate="%{text}",
            textfont={"size": 10},
        )
    )

    fig.update_layout(
        title="Sensitivity Heatmap",
        xaxis_title="KPIs",
        yaxis_title="Parameters",
        xaxis_tickangle=-45,
        width=800,
        height=600,
        paper_bgcolor="rgba(240, 240, 240, 1)",
        plot_bgcolor="rgba(240, 240, 240, 1)",
    )

    return fig


def plot_kpis(results, plot_type):
    """
    Create a horizontal bar plot comparing KPI values of different designs to a baseline.

    This function generates a plotly figure that visualizes the performance
    of various design alternatives compared to a baseline design for different Key
    Performance Indicators (KPIs).

    Parameters:
        results (pandas.DataFrame): A DataFrame containing KPI values for different designs. It should have columns for 'KPI', 'Baseline', and other design alternatives.
        plot_type (str): A string indicating the type of plot to generate. It can be either "relative" or "delta".

    Returns:
        (plotly.Figure): A Plotly Figure object representing the horizontal bar chart of KPI values, which can be displayed or further modified.
    """
    test_designs = [col for col in results.columns if col != "Baseline"]
    normalized_df = results.copy()

    if plot_type == "relative":
        for test in test_designs:
            normalized_df[test] = normalized_df[test] / normalized_df["Baseline"]
        title = "Relative Value of Designs Compared to Baseline"
        xaxis_title = "Relative Value"
    elif plot_type == "delta":
        for test in test_designs:
            normalized_df[test] = (
                normalized_df[test] - normalized_df["Baseline"]
            ) / normalized_df["Baseline"]
        title = "Delta Value of Designs Compared to Baseline"
        xaxis_title = "Delta Value"
    else:
        raise ValueError(
            "Invalid plot_type. It should be either 'relative' or 'delta'."
        )

    fig = go.Figure()

    for i, test in enumerate(test_designs):
        fig.add_trace(
            go.Bar(
                y=normalized_df.index,
                x=normalized_df[test],
                name=test,
                orientation="h",
                marker_color=hex_colors[i % len(hex_colors)],
            )
        )

    fig.update_layout(
        title=title,
        xaxis_title=xaxis_title,
        yaxis_title="KPI",
        barmode="group",
        legend_title="Design",
        width=800,
        height=600,
        paper_bgcolor="rgba(240, 240, 240, 1)",
        plot_bgcolor="rgba(240, 240, 240, 1)",
    )

    return fig


def plot_voltage_response(data):
    """
    Create a line plot of voltage response for multiple designs.

    This function generates a Plotly figure that visualizes the voltage response
    of various designs over time, specifically focusing on voltage models.

    Parameters:
        data (dict): A dictionary where keys are design names and values are dictionaries containing 'Time' and 'voltModel' data for each design. The structure should be: {design_name: {'Time': [...], 'voltModel': [...]}, ...}

    Returns:
        (plotly.Figure): A Plotly Figure object representing the line plot of dynamic KPI performance, which can be displayed or further modified.
    """
    return plot_dynamic_response(data, "voltModel")


def plot_dynamic_response(data, variable_name="voltModel"):
    """
    Create a line plot of dynamic response for multiple designs.

    This function generates a Plotly figure that visualizes the dynamic response given the variable name.

    Parameters:
        data (dict): A dictionary where keys are design names and values are dictionaries containing 'Time' and 'voltModel' data for each design. The structure should be: {design_name: {'Time': [...], 'voltModel': [...]}, ...}
        variable_name (str): The name of the variable to plot. variable_name can be one of the following: Time, voltModel, voltAnode, voltCathode, ocvModel, ocvAnode, ocvCathode, tempSurfaceModel, socModel, socAnode, socCathode, currModel, heatFluxModel.


    Returns:
        (plotly.Figure): A Plotly Figure object representing the line plot of dynamic KPI performance, which can be displayed or further modified.
    """
    fig = go.Figure()

    designs = list(data.keys())
    designs = [design for design in designs if data[design] is not None]
    num_designs = len(designs)

    for i, design in enumerate(designs):
        color = px.colors.sample_colorscale(
            hex_colors, [i / (num_designs - 1) if num_designs > 1 else 0.0]
        )[0]
        fig.add_trace(
            go.Scatter(
                x=data[design]["Time"],
                y=data[design][variable_name],
                mode="lines",
                name=design,
                line=dict(color=color, width=2),
            )
        )

    fig.update_layout(
        title="Dynamic KPI Performance",
        xaxis_title="Time",
        yaxis_title=_transform_key(variable_name),
        legend_title="Designs",
        width=800,
        height=600,
        paper_bgcolor="rgba(240, 240, 240, 1)",
        plot_bgcolor="rgba(240, 240, 240, 1)",
    )

    return fig


def plot_ocv(battery):
    """
    Create a scatter plot of OCV data for a battery.
    This function generates a plotly figure that visualizes the OCV data for a battery.

    Parameters:
        battery (dict): A dictionary containing OCV data for a battery. The structure should be: {
            "SoC": [...],
            "ChargeVoltage_V_": [...],
            "DischargeVoltage_V_": [...],
            "AnodeChargeVoltage_V_": [...]

    Returns:
        (plotly.Figure): A Plotly Figure object representing the scatter plot of OCV data, which can be displayed or further modified.
    """
    fig = go.Figure()
    x_data = battery["SoC_SoH"]
    fig.add_trace(
        go.Scatter(
            x=x_data,
            y=battery["ChargeVoltage_V_"],
            mode="lines",
            name="Charge Voltage",
            line=dict(color=hex_colors[0], width=2),
        )
    )

    fig.add_trace(
        go.Scatter(
            x=x_data,
            y=battery["DischargeVoltage_V_"],
            mode="lines",
            name="Discharge Voltage",
            line=dict(color=hex_colors[0], width=2),
        )
    )

    fig.add_trace(
        go.Scatter(
            x=x_data,
            y=battery["AnodeChargeVoltage_V_"],
            mode="lines",
            name="Anode Charge Voltage",
            line=dict(color=hex_colors[1], width=2),
        )
    )

    fig.add_trace(
        go.Scatter(
            x=x_data,
            y=battery["AnodeDischargeVoltage_V_"],
            mode="lines",
            name="Anode Discharge Voltage",
            line=dict(color=hex_colors[1], width=2),
        )
    )

    fig.add_trace(
        go.Scatter(
            x=battery["SoC"],
            y=battery["CathodeChargeVoltage_V_"],
            mode="lines",
            name="Cathode Charge Voltage",
            line=dict(color=hex_colors[2], width=2),
        )
    )

    fig.add_trace(
        go.Scatter(
            x=x_data,
            y=battery["CathodeDischargeVoltage_V_"],
            mode="lines",
            name="Cathode Discharge Voltage",
            line=dict(color=hex_colors[2], width=2),
        )
    )

    fig.update_layout(
        title="Open-Circuit Voltage (OCV) vs SoC",
        xaxis_title="SoC * SoH",
        yaxis_title="Voltage [V]",
        legend_title="Voltage Type",
        width=800,
        height=600,
        paper_bgcolor="rgba(240, 240, 240, 1)",
        plot_bgcolor="rgba(240, 240, 240, 1)",
    )

    return fig


def radar_plot(results):
    """
    Create a radar plot to visualize the performance of various designs.

    Args:
        results (pandas.DataFrame): A DataFrame containing the performance metrics of various designs.

    Returns:
        (plotly.Figure): A radar plot representing the performance of various designs.
    """
    # Define the labels for each axis
    labels = results.index  # Use the index names as labels
    results = normalize_results(results)
    # Create the radar plot
    fig = go.Figure()
    c = 0
    for i in range(len(results.columns)):
        column = results.columns[i]
        if c == len(hex_colors):
            c = 0
        fig.add_trace(
            go.Scatterpolar(
                r=results.iloc[:, i].tolist()
                + [results.iloc[0, i]],  # Add the first value to close the shape
                theta=labels.tolist() + [labels[0]],
                fill=None,
                name=column,
                line={"color": hex_colors[c]},
            )
        )
        c += 1

    # Update the layout
    fig.update_layout(
        polar=dict(radialaxis=dict(visible=True, range=[0, 2])), showlegend=True
    )

    return fig



"""
This script is a manual test to check that the authorisation works properly.

To test, run the script and just check that it gets to the end and prints out "Everything worked correctly!"
"""

from breathe_design.device_flow_auth import _DeviceFlowAuth

auth = _DeviceFlowAuth()

# try to get the token (which will require us to log in)
auth.get_token()

assert auth._token_data
assert "access_token" in auth._token_data

# check it is not expired
assert not auth._check_if_expired()

# get a new token from refresh
assert auth._get_new_token_from_refresh_token()

assert auth._token_data
assert "access_token" in auth._token_data
assert not auth._check_if_expired()

# log out
auth.logout()

print("Everything worked correctly!")



from breathe_design import api_interface as api, Cycler
from breathe_design import SingleSimulationResults, BatchSimulationResults


def test__show_service_version(use_m2m_auth):
    version = api.get_service_version()
    print(f"Service version is {version}")


def test__dynamic_analysis_I(use_m2m_auth):
    base_params = api.get_design_parameters("Molicel P45B")
    eqm_kpis = api.get_eqm_kpis("Molicel P45B")
    designs = [{"designName": "Lower NP", "NPratio": base_params["NPratio"] * 0.95}]
    baseline_capacity = eqm_kpis.capacity
    cycler = Cycler(selected_unit="C", cell_capacity=baseline_capacity)
    cycler_input = cycler.cccv(1.0, -1.0, 0.01, 4.2, 2.6, 60.0, 60.0)

    # run a single simulation
    output = api.run_sim(
        base_battery="Molicel P45B",
        cycler=cycler_input,
        designs=designs,
        initialSoC=0.5,
        initialTemperature_degC=21.0,
    )
    assert isinstance(output, SingleSimulationResults)


def test__dynamic_analysis_II(use_m2m_auth):
    design_name = "Lower NP"
    base_params = api.get_design_parameters("Molicel P45B")
    eqm_kpis = api.get_eqm_kpis("Molicel P45B")
    designs = [{"designName": "Lower NP", "NPratio": base_params["NPratio"] * 0.95}]
    baseline_capacity = eqm_kpis.capacity
    cycler = Cycler(selected_unit="C", cell_capacity=baseline_capacity)

    # define some SOC breakpoints we want to simulate
    soc_bps = [0.25, 0.5, 0.75]
    # run a batch of simulations for different SOC breakpoints
    output = api.run_sim(
        base_battery="Molicel P45B",
        cycler=cycler.cc_chg(1.0, 4.2),
        designs=designs,
        initialSoC=soc_bps,
        initialTemperature_degC=21.0,
    )

    assert isinstance(output, BatchSimulationResults)
    assert len(output) == len(soc_bps)
    # Test indexing access
    for i in range(len(soc_bps)):
        assert design_name in output.design_names
        assert output.get_dynamic_data(design_name)[0] != {}

    # Test the simulation_results_list property
    assert len(output._simulation_results) == len(soc_bps)
    for sim_result in output._simulation_results:
        assert design_name in sim_result.design_names
        assert design_name in sim_result.dynamic_data
    cycler_dict = cycler.dcir(
        I_dch=-1.0,
        t_dur=60.0,
        t_rest_before=3.0,
        t_rest_after=200.0,
        v_min=2.5,
        v_max=4.2,
    )
    # DCIR
    output = api.run_sim(
        base_battery="Molicel P45B",
        cycler=cycler_dict,
        designs=designs,
        initialSoC=0.5,
        initialTemperature_degC=21.0,
    )

    # only one initial SOC and temperature, so we expect a SingleSimulationResults back
    assert isinstance(output, SingleSimulationResults)
    assert design_name in output.design_names
    assert design_name in output.dynamic_data

    # Test the dynamic_kpis method from the results handler
    dynamic_kpis = output.dynamic_kpis("DCIR")
    assert dynamic_kpis is not None



import pytest
from breathe_design.format import (
    validate_material,
    validate_assembly_type,
    validate_format_parameter,
    define_cuboid_format,
    define_cylinder_format,
    VALID_MATERIALS,
    VALID_ASSEMENT_TYPES,
)
from breathe_design.parameter_validator import parameter_validator


class TestFormatValidation:
    """Test cases for format validation functions."""

    def test_valid_materials_constant(self):
        """Test that VALID_MATERIALS contains expected materials."""
        assert "steel" in VALID_MATERIALS
        assert "aluminum" in VALID_MATERIALS
        assert "pouch" in VALID_MATERIALS
        assert len(VALID_MATERIALS) == 3

    def test_valid_assembly_types_constant(self):
        """Test that VALID_ASSEMENT_TYPES contains expected types."""
        assert "stacked" in VALID_ASSEMENT_TYPES
        assert "wound" in VALID_ASSEMENT_TYPES
        assert "zFolded" in VALID_ASSEMENT_TYPES
        assert len(VALID_ASSEMENT_TYPES) == 3

    def test_validate_material_valid(self):
        """Test validation of valid materials."""
        for material in VALID_MATERIALS:
            # Should not raise an exception
            validate_material(material)

    def test_validate_material_invalid(self):
        """Test validation fails for invalid materials."""
        invalid_materials = ["plastic", "wood", "ceramic", "invalid"]

        for material in invalid_materials:
            with pytest.raises(ValueError, match="Invalid material"):
                validate_material(material)

    def test_validate_assembly_type_valid(self):
        """Test validation of valid assembly types."""
        for assembly_type in VALID_ASSEMENT_TYPES:
            # Should not raise an exception
            validate_assembly_type(assembly_type)

    def test_validate_assembly_type_invalid(self):
        """Test validation fails for invalid assembly types."""
        invalid_types = ["rolled", "folded", "twisted", "invalid"]

        for assembly_type in invalid_types:
            with pytest.raises(ValueError, match="Invalid assembly type"):
                validate_assembly_type(assembly_type)

    def test_validate_format_parameter_valid(self):
        """Test validation of valid format parameters."""
        # Should not raise exceptions
        validate_format_parameter("housingThickness_mm", 1.0)
        validate_format_parameter("length_mm", 100.0)
        validate_format_parameter("diameter_mm", 18.0)

    def test_validate_format_parameter_out_of_bounds_high(self):
        """Test format parameter validation fails for values above maximum."""
        with pytest.raises(ValueError, match="above maximum allowed value"):
            validate_format_parameter("housingThickness_mm", 15.0)  # Max is 10.0

    def test_validate_format_parameter_out_of_bounds_low(self):
        """Test format parameter validation fails for values below minimum."""
        with pytest.raises(ValueError, match="below minimum allowed value"):
            validate_format_parameter("housingThickness_mm", 0.05)  # Min is 0.1

    def test_validate_format_parameter_none(self):
        """Test format parameter validation fails for None values."""
        with pytest.raises(ValueError, match="cannot be None"):
            validate_format_parameter("housingThickness_mm", None)

    def test_validate_format_parameter_non_numeric(self):
        """Test format parameter validation fails for non-numeric values."""
        with pytest.raises(ValueError, match="must be numeric"):
            validate_format_parameter("housingThickness_mm", "not_a_number")

    def test_validate_format_parameter_unknown_parameter(self):
        """Test that unknown parameters are silently skipped."""
        # Should not raise an exception
        validate_format_parameter("unknown_parameter", 42.0)

    def test_define_cuboid_format_valid(self):
        """Test creation of valid cuboid format."""
        result = define_cuboid_format(
            name="Test Cuboid",
            material="aluminum",
            housingThickness_mm=1.0,
            length_mm=100.0,
            width_mm=50.0,
            thickness_mm=10.0,
            assemblyType="stacked",
        )

        assert result["shape"] == "cuboid"
        assert result["name"] == "Test Cuboid"
        assert result["material"] == "aluminum"
        assert result["housingThickness_mm"] == 1.0
        assert result["length_mm"] == 100.0
        assert result["width_mm"] == 50.0
        assert result["thickness_mm"] == 10.0
        assert result["assemblyType"] == "stacked"

        # Check that numeric values are converted to float
        assert isinstance(result["housingThickness_mm"], float)
        assert isinstance(result["length_mm"], float)

    def test_define_cuboid_format_invalid_material(self):
        """Test cuboid creation fails with invalid material."""
        with pytest.raises(ValueError, match="Invalid material"):
            define_cuboid_format(
                name="Invalid Material",
                material="plastic",
                housingThickness_mm=1.0,
                length_mm=100.0,
                width_mm=50.0,
                thickness_mm=10.0,
                assemblyType="stacked",
            )

    def test_define_cuboid_format_invalid_assembly_type(self):
        """Test cuboid creation fails with invalid assembly type."""
        with pytest.raises(ValueError, match="Invalid assembly type"):
            define_cuboid_format(
                name="Invalid Assembly",
                material="aluminum",
                housingThickness_mm=1.0,
                length_mm=100.0,
                width_mm=50.0,
                thickness_mm=10.0,
                assemblyType="invalid_type",
            )

    def test_define_cuboid_format_out_of_bounds_parameter(self):
        """Test cuboid creation fails with out-of-bounds parameter."""
        with pytest.raises(ValueError, match="above maximum allowed value"):
            define_cuboid_format(
                name="Too Thick",
                material="aluminum",
                housingThickness_mm=15.0,  # Above maximum
                length_mm=100.0,
                width_mm=50.0,
                thickness_mm=10.0,
                assemblyType="stacked",
            )

    def test_define_cylinder_format_valid(self):
        """Test creation of valid cylinder format."""
        result = define_cylinder_format(
            name="Test Cylinder",
            material="steel",
            housingThickness_mm=0.5,
            diameter_mm=18.0,
            height_mm=65.0,
            innerDiameter_mm=17.0,
        )

        assert result["shape"] == "cylinder"
        assert result["name"] == "Test Cylinder"
        assert result["material"] == "steel"
        assert result["housingThickness_mm"] == 0.5
        assert result["diameter_mm"] == 18.0
        assert result["height_mm"] == 65.0
        assert result["innerDiameter_mm"] == 17.0

        # Check that numeric values are converted to float
        assert isinstance(result["housingThickness_mm"], float)
        assert isinstance(result["diameter_mm"], float)

    def test_define_cylinder_format_invalid_material(self):
        """Test cylinder creation fails with invalid material."""
        with pytest.raises(ValueError, match="Invalid material"):
            define_cylinder_format(
                name="Invalid Material",
                material="plastic",
                housingThickness_mm=0.5,
                diameter_mm=18.0,
                height_mm=65.0,
                innerDiameter_mm=17.0,
            )

    def test_define_cylinder_format_invalid_constraints(self):
        """Test cylinder creation fails when inner diameter >= outer diameter."""
        with pytest.raises(ValueError, match="must be less than outer diameter"):
            define_cylinder_format(
                name="Invalid Cylinder",
                material="steel",
                housingThickness_mm=0.5,
                diameter_mm=18.0,
                height_mm=65.0,
                innerDiameter_mm=18.0,  # Equal to outer diameter
            )

        with pytest.raises(ValueError, match="must be less than outer diameter"):
            define_cylinder_format(
                name="Invalid Cylinder 2",
                material="steel",
                housingThickness_mm=0.5,
                diameter_mm=18.0,
                height_mm=65.0,
                innerDiameter_mm=19.0,  # Greater than outer diameter
            )

    def test_define_cylinder_format_out_of_bounds_parameter(self):
        """Test cylinder creation fails with out-of-bounds parameter."""
        with pytest.raises(ValueError, match="above maximum allowed value"):
            define_cylinder_format(
                name="Too Thick",
                material="steel",
                housingThickness_mm=15.0,  # Above maximum
                diameter_mm=18.0,
                height_mm=65.0,
                innerDiameter_mm=17.0,
            )

    def test_parameter_validator_format_validation(self):
        """Test that the unified parameter validator can validate formats."""
        # parameter_validator should be the unified parameter validator
        assert parameter_validator is not None

        # Test that it has the expected methods
        assert hasattr(parameter_validator, "validate_format")
        assert hasattr(parameter_validator, "validate_formats")
        assert hasattr(parameter_validator, "validate_design")
        assert hasattr(parameter_validator, "validate_designs")

    def test_unified_validator_integration(self):
        """Test integration with the unified parameter validator."""
        # Create a format and validate it using the unified validator
        format_dict = {
            "shape": "cuboid",
            "name": "Integration Test",
            "material": "aluminum",
            "housingThickness_mm": 1.0,
            "length_mm": 100.0,
            "width_mm": 50.0,
            "thickness_mm": 10.0,
            "assemblyType": "stacked",
        }

        result = parameter_validator.validate_format(format_dict)
        assert result["shape"] == "cuboid"
        assert result["material"] == "aluminum"

    def test_numeric_parameter_type_conversion(self):
        """Test that define functions convert integers to floats."""
        # Test with integer inputs
        cuboid = define_cuboid_format(
            name="Integer Test",
            material="aluminum",
            housingThickness_mm=1,  # Integer
            length_mm=100,  # Integer
            width_mm=50,  # Integer
            thickness_mm=10,  # Integer
            assemblyType="stacked",
        )

        # All numeric parameters should be floats
        assert isinstance(cuboid["housingThickness_mm"], float)
        assert isinstance(cuboid["length_mm"], float)
        assert isinstance(cuboid["width_mm"], float)
        assert isinstance(cuboid["thickness_mm"], float)

        # Values should be preserved
        assert cuboid["housingThickness_mm"] == 1.0
        assert cuboid["length_mm"] == 100.0

    def test_boundary_values(self):
        """Test format validation with boundary values."""
        # Test minimum boundary values
        min_cuboid = define_cuboid_format(
            name="Min Values",
            material="aluminum",
            housingThickness_mm=0.1,  # Minimum
            length_mm=1.0,  # Minimum
            width_mm=1.0,  # Minimum
            thickness_mm=1.0,  # Minimum
            assemblyType="stacked",
        )
        assert min_cuboid["housingThickness_mm"] == 0.1

        # Test maximum boundary values
        max_cuboid = define_cuboid_format(
            name="Max Values",
            material="aluminum",
            housingThickness_mm=10.0,  # Maximum
            length_mm=1000.0,  # Maximum
            width_mm=1000.0,  # Maximum
            thickness_mm=100.0,  # Maximum
            assemblyType="stacked",
        )
        assert max_cuboid["housingThickness_mm"] == 10.0

    def test_error_message_specificity(self):
        """Test that error messages are specific and helpful."""
        # Test specific parameter name in error message
        with pytest.raises(ValueError) as exc_info:
            validate_format_parameter("housingThickness_mm", 15.0)

        assert "housingThickness_mm" in str(exc_info.value)
        assert "15.0" in str(exc_info.value)
        assert "10.0" in str(exc_info.value)  # Maximum value

        # Test material error message
        with pytest.raises(ValueError) as exc_info:
            validate_material("plastic")

        assert "plastic" in str(exc_info.value)
        assert "aluminum" in str(exc_info.value) or "steel" in str(
            exc_info.value
        )  # Valid options shown

    def test_unified_validator_format_list_validation(self):
        """Test that the unified validator can validate lists of formats."""
        formats = [
            {
                "shape": "cuboid",
                "name": "Format 1",
                "material": "aluminum",
                "housingThickness_mm": 1.0,
                "length_mm": 100.0,
            },
            {
                "shape": "cylinder",
                "name": "Format 2",
                "material": "steel",
                "housingThickness_mm": 0.5,
                "diameter_mm": 18.0,
                "height_mm": 65.0,
                "innerDiameter_mm": 17.0,
            },
        ]

        result = parameter_validator.validate_formats(formats)
        assert len(result) == 2
        assert result[0]["shape"] == "cuboid"
        assert result[1]["shape"] == "cylinder"

    def test_unified_validator_format_parameter_bounds(self):
        """Test that the unified validator uses the same parameter bounds as format functions."""
        # Get bounds from unified validator
        bounds = parameter_validator.get_parameter_bounds()

        # Test that format-specific parameters exist
        assert "housingThickness_mm" in bounds
        assert "length_mm" in bounds
        assert "diameter_mm" in bounds

        # Test that bounds match expected values
        assert bounds["housingThickness_mm"] == [0.1, 10.0]
        assert bounds["length_mm"] == [1.0, 1000.0]

    def test_unified_validator_format_validation_consistency(self):
        """Test that unified validator and format functions give consistent results."""
        # Test a format that should pass both validations
        valid_format = {
            "shape": "cuboid",
            "name": "Consistency Test",
            "material": "aluminum",
            "housingThickness_mm": 2.0,
            "length_mm": 150.0,
            "width_mm": 75.0,
            "thickness_mm": 15.0,
            "assemblyType": "stacked",
        }

        # Both should succeed
        unified_result = parameter_validator.validate_format(valid_format)
        format_function_result = define_cuboid_format(
            name="Consistency Test",
            material="aluminum",
            housingThickness_mm=2.0,
            length_mm=150.0,
            width_mm=75.0,
            thickness_mm=15.0,
            assemblyType="stacked",
        )

        # Key fields should match
        assert unified_result["material"] == format_function_result["material"]
        assert (
            unified_result["housingThickness_mm"]
            == format_function_result["housingThickness_mm"]
        )

    def test_unified_validator_format_error_consistency(self):
        """Test that unified validator and format functions give consistent errors."""
        # Test a parameter that should fail in both
        invalid_thickness = 15.0  # Above maximum of 10.0

        # Both should fail with similar error messages
        with pytest.raises(ValueError, match="above maximum allowed value"):
            parameter_validator.validate_format(
                {
                    "shape": "cuboid",
                    "material": "aluminum",
                    "housingThickness_mm": invalid_thickness,
                }
            )

        with pytest.raises(ValueError, match="above maximum allowed value"):
            define_cuboid_format(
                name="Error Test",
                material="aluminum",
                housingThickness_mm=invalid_thickness,
                length_mm=100.0,
                width_mm=50.0,
                thickness_mm=10.0,
                assemblyType="stacked",
            )



import pytest
import os
from breathe_design.cycler import Cycler


class TestCycler:
    """Test cases for the refactored Cycler class."""

    def test_cycler_initialization_A_units(self):
        """Test Cycler initialization with Ampere units."""
        cycler = Cycler("A", 3.0)
        assert cycler._selected_unit == "A"
        assert cycler._cell_capacity == 3.0
        assert cycler._protocols is not None
        assert isinstance(cycler._protocols, dict)

    def test_cycler_initialization_C_units(self):
        """Test Cycler initialization with C-rate units."""
        cycler = Cycler("C", 2.5)
        assert cycler._selected_unit == "C"
        assert cycler._cell_capacity == 2.5

    def test_protocols_loaded_from_json(self):
        """Test that protocols are loaded from JSON file."""
        cycler = Cycler("A", 3.0)

        # Check that known protocols exist
        assert "CC_CHG" in cycler._protocols
        assert "CC_DCH" in cycler._protocols
        assert "DCIR" in cycler._protocols
        assert "CCCV" in cycler._protocols
        assert "RateCap" in cycler._protocols

        # Check protocol structure
        cc_chg = cycler._protocols["CC_CHG"]
        assert "I_chg" in cc_chg
        assert "V_max" in cc_chg
        assert isinstance(cc_chg["I_chg"], list)
        assert len(cc_chg["I_chg"]) == 2  # [min, max]

    def test_cc_chg_valid_parameters(self):
        """Test CC_CHG protocol with valid parameters."""
        cycler = Cycler("A", 3.0)
        result = cycler.cc_chg(I_chg=1.0, V_max=4.2)

        assert result["cycle_type"] == "CC_CHG"
        assert result["selected_unit"] == "A"
        assert result["control_parameters"]["I_chg"] == 1.0
        assert result["control_parameters"]["V_max"] == 4.2

    def test_cc_chg_c_rate_conversion(self):
        """Test CC_CHG protocol with C-rate units."""
        cycler = Cycler("C", 3.0)  # 3 Ah capacity
        result = cycler.cc_chg(I_chg=1.0, V_max=4.2)  # 1C = 3A

        assert result["cycle_type"] == "CC_CHG"
        assert result["selected_unit"] == "C"
        assert result["control_parameters"]["I_chg"] == 1.0  # Input value preserved
        assert result["control_parameters"]["V_max"] == 4.2

    def test_cc_chg_out_of_bounds_current(self):
        """Test CC_CHG fails with out-of-bounds current."""
        cycler = Cycler("A", 3.0)

        with pytest.raises(ValueError, match="Invalid value for I_chg"):
            cycler.cc_chg(I_chg=600.0, V_max=4.2)  # Above maximum

    def test_cc_chg_out_of_bounds_voltage(self):
        """Test CC_CHG fails with out-of-bounds voltage."""
        cycler = Cycler("A", 3.0)

        with pytest.raises(ValueError, match="Invalid value for V_max"):
            cycler.cc_chg(I_chg=1.0, V_max=5.0)  # Above maximum

    def test_cc_chg_missing_parameter(self):
        """Test CC_CHG fails with missing required parameter."""
        cycler = Cycler("A", 3.0)

        with pytest.raises(TypeError, match="missing 1 required positional argument"):
            cycler.cc_chg(I_chg=1.0)  # Missing V_max

    def test_cc_chg_none_parameter(self):
        """Test CC_CHG fails with None parameter."""
        cycler = Cycler("A", 3.0)

        with pytest.raises(ValueError, match="Missing required parameter"):
            cycler.cc_chg(I_chg=None, V_max=4.2)

    def test_cc_dch_valid_parameters(self):
        """Test CC_DCH protocol with valid parameters."""
        cycler = Cycler("A", 3.0)
        result = cycler.cc_dch(I_dch=-2.0, V_min=2.8)

        assert result["cycle_type"] == "CC_DCH"
        assert result["control_parameters"]["I_dch"] == -2.0
        assert result["control_parameters"]["V_min"] == 2.8

    def test_dcir_valid_parameters(self):
        """Test DCIR protocol with valid parameters."""
        cycler = Cycler("A", 3.0)
        result = cycler.dcir(
            I_dch=-5.0,
            t_dur=10.0,
            t_rest_before=5.0,
            t_rest_after=5.0,
            v_min=2.8,
            v_max=4.2,
        )

        assert result["cycle_type"] == "DCIR"
        assert result["control_parameters"]["I_dch"] == -5.0
        assert result["control_parameters"]["t_dur"] == 10.0
        assert result["control_parameters"]["t_rest_before"] == 5.0
        assert result["control_parameters"]["t_rest_after"] == 5.0
        assert result["control_parameters"]["v_min"] == 2.8
        assert result["control_parameters"]["v_max"] == 4.2

    def test_cccv_valid_parameters(self):
        """Test CCCV protocol with valid parameters."""
        cycler = Cycler("A", 3.0)
        result = cycler.cccv(
            I_chg=1.0,
            I_dch=-2.0,
            I_cut=0.1,
            V_max=4.2,
            V_min=2.8,
            t_restC=300.0,
            t_restD=300.0,
        )

        assert result["cycle_type"] == "CCCV"
        assert result["control_parameters"]["I_chg"] == 1.0
        assert result["control_parameters"]["I_dch"] == -2.0
        assert result["control_parameters"]["I_cut"] == 0.1

    def test_rate_cap_valid_parameters(self):
        """Test RateCap protocol with valid parameters."""
        cycler = Cycler("A", 3.0)
        result = cycler.rate_cap(
            I_chg=1.0, I_dch=-2.0, I_cut=0.1, V_max=4.2, V_min=2.8, t_restC=300.0
        )

        # rate_cap method should correctly set cycle_type to "RateCap"
        assert result["cycle_type"] == "RateCap"
        assert result["control_parameters"]["I_chg"] == 1.0

    def test_invalid_cycle_type(self):
        """Test that invalid cycle type raises error."""
        cycler = Cycler("A", 3.0)

        # Manually create invalid cycler dict to test _build_cycler_dict
        invalid_cycler = {
            "cycle_type": "INVALID_TYPE",
            "selected_unit": "A",
            "control_parameters": {},
        }

        with pytest.raises(ValueError, match="Invalid cycle type"):
            cycler._build_cycler_dict(invalid_cycler)

    def test_c_rate_bounds_checking(self):
        """Test that C-rate values are converted for bounds checking."""
        cycler = Cycler("C", 3.0)  # 3 Ah capacity

        # 200C would be 600A, which is above the 500A limit
        with pytest.raises(ValueError, match="Invalid value for I_chg"):
            cycler.cc_chg(I_chg=200.0, V_max=4.2)  # 200C = 600A > 500A max

    def test_c_rate_valid_within_bounds(self):
        """Test that valid C-rate values pass bounds checking."""
        cycler = Cycler("C", 3.0)  # 3 Ah capacity

        # 1C = 3A, which is well within bounds
        result = cycler.cc_chg(I_chg=1.0, V_max=4.2)
        assert result["control_parameters"]["I_chg"] == 1.0

    def test_protocols_structure(self):
        """Test that protocols have the expected flattened structure."""
        cycler = Cycler("A", 3.0)

        for protocol_name, protocol in cycler._protocols.items():
            assert isinstance(protocol, dict)

            # Each parameter should map to [min, max] list
            for param_name, bounds in protocol.items():
                assert isinstance(bounds, list)
                assert len(bounds) == 2
                min_val, max_val = bounds

                # min_val can be None or numeric
                if min_val is not None:
                    assert isinstance(min_val, (int, float))

                # max_val can be None or numeric
                if max_val is not None:
                    assert isinstance(max_val, (int, float))

    def test_load_protocols_file_not_found(self):
        """Test that missing protocols file raises appropriate error."""
        # This is harder to test without mocking, but we can test the error message format
        # by examining the code behavior when file doesn't exist

        # Create a temporary Cycler class that tries to load from non-existent file
        original_dirname = os.path.dirname

        def mock_dirname(path):
            return "/non/existent/path"

        # Temporarily replace os.path.dirname
        os.path.dirname = mock_dirname
        try:
            with pytest.raises(
                FileNotFoundError, match="Could not find cycling protocols file"
            ):
                Cycler("A", 3.0)
        finally:
            # Restore original function
            os.path.dirname = original_dirname

    def test_protocols_json_content_validation(self):
        """Test that protocols loaded from JSON have valid content."""
        cycler = Cycler("A", 3.0)

        # Test CC_CHG protocol specifically
        cc_chg = cycler._protocols["CC_CHG"]
        assert cc_chg["I_chg"] == [0.0, 500.0]
        assert cc_chg["V_max"] == [2.5, 4.5]

        # Test DCIR protocol has all expected parameters
        dcir = cycler._protocols["DCIR"]
        expected_dcir_params = [
            "I_dch",
            "t_dur",
            "t_rest_before",
            "t_rest_after",
            "v_min",
            "v_max",
        ]
        for param in expected_dcir_params:
            assert param in dcir
            assert isinstance(dcir[param], list)
            assert len(dcir[param]) == 2

    def test_default_cycler_structure(self):
        """Test that default cycler structure is properly initialized."""
        cycler = Cycler("A", 3.0)

        # Check default cycler structure
        default = cycler._default_cycler
        assert default["cycle_type"] is None
        assert default["selected_unit"] == "A"
        assert default["control_parameters"] == {}

    def test_parameter_validation_edge_cases(self):
        """Test edge cases in parameter validation."""
        cycler = Cycler("A", 3.0)

        # Test minimum boundary values
        result = cycler.cc_chg(I_chg=0.0, V_max=2.5)  # Minimum values
        assert result["control_parameters"]["I_chg"] == 0.0
        assert result["control_parameters"]["V_max"] == 2.5

        # Test maximum boundary values
        result = cycler.cc_chg(I_chg=500.0, V_max=4.5)  # Maximum values
        assert result["control_parameters"]["I_chg"] == 500.0
        assert result["control_parameters"]["V_max"] == 4.5

    def test_docstring_accuracy(self):
        """Test that method docstrings accurately describe the functionality."""
        cycler = Cycler("A", 3.0)

        # Check that CC_CHG method returns what docstring claims
        result = cycler.cc_chg(I_chg=1.0, V_max=4.2)
        assert isinstance(result, dict), "cc_chg should return dict as documented"

        # Check that DCIR method returns what docstring claims
        result = cycler.dcir(
            I_dch=-5.0,
            t_dur=10.0,
            t_rest_before=5.0,
            t_rest_after=5.0,
            v_min=2.8,
            v_max=4.2,
        )
        assert isinstance(result, dict), "dcir should return dict as documented"



from breathe_design.api_utils import (
    map_fields_in_place,
    make_design_names_map,
    map_design_names__human_to_machine,
)


def test__make_design_names_map():
    map_h2m, map_m2h = make_design_names_map([{"designName": "Unsafe Name %!?"}])

    assert map_h2m == {"Unsafe Name %!?": "design_0"}
    assert map_m2h == {"design_0": "Unsafe Name %!?"}


def test__map_fields_in_place():
    """Test the inverse mapping from safe machine names to the human unsafe ones.

    This is the mpapping that is done on the results from the server
    """
    unsafe_name = "Unsafe Name %!?"
    map_h2m, map_m2h = make_design_names_map([{"designName": unsafe_name}])

    results = {}
    results["KPIs"] = {"Baseline": 0, unsafe_name: (1, 2, 3)}

    # map fowards to the safe names
    map_fields_in_place(results, "KPIs", map_h2m)
    assert results["KPIs"] == {"Baseline": 0, "design_0": (1, 2, 3)}

    # map back (as if they came back from the server)
    map_fields_in_place(results, "KPIs", map_m2h)
    assert results["KPIs"] == {"Baseline": 0, unsafe_name: (1, 2, 3)}


def test__map_design_names__human_to_machine():
    """Test the mapping from unsafe designNames to safe ones.

    This is the mapping that is performed on the designs before sending to the server
    """
    unsafe_name = "Unsafe Name %!?"
    designs = [{"designName": unsafe_name}]
    map_h2m, map_m2h = make_design_names_map(designs)

    # map fowards to the safe names
    designs_machine = map_design_names__human_to_machine(designs, map_h2m)
    assert designs_machine == [{"designName": "design_0"}]



import os
from slack_sdk import WebClient
import re
import logging
from breathe_utils import logging as breathe_logging

channel = os.getenv("SLACK_CHANNEL")
token = os.getenv("SLACK_TOKEN")

breathe_logging.setup_logging("log.log", append=False)
_logger = logging.getLogger(__name__)

# Read the changelog file
try:
    with open("CHANGELOG.md", "r") as file:
        changelog_content = file.read()
except:  # noqa: E722
    try:
        with open("Changelog.md", "r") as file:
            changelog_content = file.read()
    except:  # noqa: E722
        _logger.error("No changelog file found")


# Define the pattern to match release entries
pattern = r"## \[(.*?)\] - (\d{2}/\d{2}/\d{4})\n(.*?)(?=\s*## \[|\Z)"

# Find all release entries
matches = re.findall(pattern, changelog_content, re.DOTALL)

if matches:
    # Get the latest release
    latest_release = matches[0]
    version = latest_release[0]
    date = latest_release[1]
    changes = latest_release[2].strip()

    ci_project_name = os.getenv("CI_PROJECT_TITLE")

    # Format the details into a Slack message
    slack_message = f"New Release of {ci_project_name} (v{version}):\n\n{changes}"

    # Send the Slack message or print it for testing
    print(slack_message)
else:
    print("No releases found in the changelog.")


# Set up a WebClient with the Slack OAuth token
client = WebClient(token=token)

# Send a message
client.chat_postMessage(channel=channel, text=slack_message, username="Breathe DevOps")
